# cython: language_level=3
# cython: profile=True
# Time-stamp: <2024-10-14 17:04:27 Tao Liu>

"""Module description:

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""
# ------------------------------------
# python modules
# ------------------------------------
import cython
from MACS3.Signal.Prob import pnorm2
from MACS3.Signal.BedGraph import bedGraphTrackI
from MACS3.Signal.Region import Regions
from MACS3.Utilities.Logger import logging

logger = logging.getLogger(__name__)
debug = logger.debug
info = logger.info

# ------------------------------------
# Misc functions
# ------------------------------------


@cython.inline
@cython.cfunc
def get_weighted_density(x: cython.int, m: cython.float, v:
                         cython.float, w: cython.float) -> cython.float:
    """Description:

    parameters:
      1. x: the observed value
      2. m: the mean of gaussian
      3. v: the variance of the gaussian
      4. w: the weight
    return value:
    """
    return w * pnorm2(float(x), m, v)


# ------------------------------------
# Classes
# ------------------------------------
# ------------------------------------
# public functions
# ------------------------------------


@cython.ccall
def generate_weight_mapping(fraglen_list: list, means: list, stddevs:
                            list, min_frag_p: cython.float = 0.001) -> list:
    """Generate weights for each fragment length in short, mono, di,
    and tri-signals track

    return: list of four dictionaries, with key as fraglen and value
    as the weight.

    ret[0] -- dictionary for short
    ret[1] -- dictionary for mono
    ret[2] -- dictionary for di
    ret[3] -- dictionary for tri

    """

    ret_mapping: list
    fl: cython.int
    m_s: cython.float
    m_m: cython.float
    m_d: cython.float
    m_t: cython.float
    v_s: cython.float
    v_m: cython.float
    v_d: cython.float
    v_t: cython.float
    p_s: cython.float
    p_m: cython.float
    p_d: cython.float
    p_t: cython.float
    s: cython.float
    i: cython.int

    assert len(means) == 4
    assert len(stddevs) == 4
    [m_s, m_m, m_d, m_t] = means
    [v_s, v_m, v_d, v_t] = [x**2 for x in stddevs]
    ret_mapping = [{}, {}, {}, {}]
    for i in range(len(fraglen_list)):
        fl = fraglen_list[i]
        p_s = pnorm2(float(fl), m_s, v_s)
        p_m = pnorm2(float(fl), m_m, v_m)
        p_d = pnorm2(float(fl), m_d, v_d)
        p_t = pnorm2(float(fl), m_t, v_t)
        s = p_s + p_m + p_d + p_t
        if p_s < min_frag_p and p_m < min_frag_p and p_d < min_frag_p and p_t < min_frag_p:
            # we exclude the fragment which can't be assigned to
            # short, mono, di-nuc, and tri-nuc (likelihood <
            # min_frag_p, default:0.001) Normally this fragment is too
            # large. We exclude these fragment by setting all weights
            # to zero.
            debug(f"The fragment length {fl} can't be assigned to either distribution so will be excluded!")
            ret_mapping[0][fl] = 0
            ret_mapping[1][fl] = 0
            ret_mapping[2][fl] = 0
            ret_mapping[3][fl] = 0
            continue
        ret_mapping[0][fl] = p_s / s
        ret_mapping[1][fl] = p_m / s
        ret_mapping[2][fl] = p_d / s
        ret_mapping[3][fl] = p_t / s
    return ret_mapping


@cython.ccall
def generate_digested_signals(petrack, weight_mapping: list) -> list:
    """Generate digested pileup signals (four tracks) using weight mapping

    return: list of four signals in dictionary, with key as chromosome name and value as a p-v array.
            ret[0] -- dictionary for short
            ret[1] -- dictionary for mono
            ret[2] -- dictionary for di
            ret[3] -- dictionary for tri
    """
    ret_digested_signals: list
    ret_bedgraphs: list
    bdg: object
    i: int
    certain_signals: dict
    chrom: bytes

    ret_digested_signals = petrack.pileup_bdg_hmmr(weight_mapping)
    ret_bedgraphs = []
    for i in range(4):          # yes I hardcoded 4!
        certain_signals = ret_digested_signals[i]
        bdg = bedGraphTrackI()
        for chrom in sorted(certain_signals.keys()):
            bdg.add_chrom_data_PV(chrom, certain_signals[chrom])
        ret_bedgraphs.append(bdg)
    return ret_bedgraphs


@cython.ccall
def extract_signals_from_regions(signals: list, regions, binsize:
                                 cython.int = 10, hmm_type: str = 'gaussian') -> list:
    # we will take regions in peaks, create a bedGraphTrackI with
    # binned regions in peaks, then let them overlap with signals to
    # create a list (4) of value arrays.
    #
    extracted_data: list
    extracted_len: list
    extracted_positions: list
    signaltrack: object
    regionsbdg: object
    i: cython.int
    c: cython.int
    counter: cython.int
    prev_c: cython.int
    ret_training_data: list
    ret_training_lengths: list
    ret_training_bins: list

    regionsbdg = _make_bdg_of_bins_from_regions(regions, binsize)
    debug('#      extract_signals_from_regions: regionsbdg completed')
    # now, let's overlap
    extracted_positions = []
    extracted_data = []
    extracted_len = []
    for signaltrack in signals: # four signal tracks
        # signaltrack is bedGraphTrackI object
        [positions, values, lengths] = signaltrack.extract_value_hmmr(regionsbdg)
        extracted_positions.append(positions)
        extracted_data.append(values)
        extracted_len.append(lengths)
    positions = []
    values = []
    lengths = []
    debug('#      extract_signals_from_regions: extracted positions, data, len')
    ret_training_bins = []
    ret_training_data = []
    ret_training_lengths = []

    nn = len(extracted_data[0])
    assert nn > 0
    assert nn == len(extracted_data[1])
    assert nn == len(extracted_data[2])
    assert nn == len(extracted_data[3])
    counter = 0
    prev_c = extracted_len[0][0]
    c = 0
    if hmm_type == "gaussian":
        for i in range(nn):
            ret_training_bins.append(extracted_positions[0][i])
            ret_training_data.append(
                [max(0.0001, extracted_data[0][i]),
                 max(0.0001, extracted_data[1][i]),
                 max(0.0001, extracted_data[2][i]),
                 max(0.0001, extracted_data[3][i])])
            c = extracted_len[0][i]
            if counter != 0 and c != prev_c:
                ret_training_lengths.append(counter)
                counter = 0
            prev_c = c
            counter += 1
        debug('#      extract_signals_from_regions: ret_training bins, data, lengths - gaussian')
    # poisson can only take int values as input
    if hmm_type == "poisson":
        for i in range(nn):
            ret_training_bins.append(extracted_positions[0][i])
            ret_training_data.append(
                [int(max(0.0001, extracted_data[0][i])),
                 int(max(0.0001, extracted_data[1][i])),
                 int(max(0.0001, extracted_data[2][i])),
                 int(max(0.0001, extracted_data[3][i]))])
            c = extracted_len[0][i]
            if counter != 0 and c != prev_c:
                ret_training_lengths.append(counter)
                counter = 0
            prev_c = c
            counter += 1
        debug('#      extract_signals_from_regions: ret_training bins, data, lengths - poisson')
    # last region
    ret_training_lengths.append(counter)
    assert sum(ret_training_lengths) == len(ret_training_data)
    assert len(ret_training_bins) == len(ret_training_data)
    return [ret_training_bins, ret_training_data, ret_training_lengths]


@cython.cfunc
def _make_bdg_of_bins_from_regions(regions, binsize: cython.int):
    # this function will return a BedGraphTrackI object
    regionsbdg: object
    n: cython.long
    chrom: bytes
    ps: list
    s: cython.int
    e: cython.int
    tmp_p: cython.int
    mark_bin: cython.int
    i: cython.int
    r: cython.int

    assert isinstance(regions, Regions)

    regionsbdg = bedGraphTrackI(baseline_value=-100)

    n = 0
    # here we convert peaks from a PeakIO to BedGraph object with a
    # given binsize.
    mark_bin = 1  # this is to mark the continuous bins in the same region, it will increase by one while moving to the next region
    for chrom in sorted(regions.get_chr_names()):
        tmp_p = 0  # this is to make gap in bedgraph for not covered regions.
        ps = regions[chrom]
        for i in range(len(ps)):
            # for each region
            s = ps[i][0]
            e = ps[i][1]
            # make bins, no need to be too accurate...
            s = s//binsize*binsize
            e = e//binsize*binsize
            # tmp_n = int((e - s)/binsize)
            for r in range(s, e, binsize):
                tmp_s = r
                tmp_e = r + binsize
                if tmp_s > tmp_p:
                    regionsbdg.add_loc_wo_merge(chrom, tmp_p, tmp_s, 0)  # the gap
                regionsbdg.add_loc_wo_merge(chrom, tmp_s, tmp_e, mark_bin)  # the value we put in the bin bedgraph is the number of bins in this region
                n += 1
                tmp_p = tmp_e
            # end of region, we change the mark_bin
            mark_bin += 1
    # we do not merge regions in regionsbdg object so each bin will be separated.
    debug(f"added {n} bins")
    return regionsbdg



# cython: language_level=3
# cython: profile=True
# cython: linetrace=True
# Time-stamp: <2025-11-12 22:14:22 Tao Liu>

"""Input parsers used across MACS3 for reading alignment-like formats.

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

# ------------------------------------
# python modules
# ------------------------------------
import struct
from struct import unpack
from re import findall
import gzip
import io
import sys

import cython
from cython.cimports.cpython import bool

from cython.cimports.libc.stdlib import atoi

from MACS3.Utilities.Constants import READ_BUFFER_SIZE
from MACS3.Signal.FixWidthTrack import FWTrack
from MACS3.Signal.PairedEndTrack import PETrackI, PETrackII
from MACS3.Utilities.Logger import logging

logger = logging.getLogger(__name__)
debug = logger.debug
info = logger.info
warn = logger.warn
# ------------------------------------
# Other modules
# ------------------------------------

if sys.byteorder == "little":
    endian_prefix = "<"
elif sys.byteorder == "big":
    endian_prefix = ">"
else:
    raise Exception("Byteorder should be either big or little-endian.")

# ------------------------------------
# Misc functions
# ------------------------------------


@cython.ccall
def guess_parser(fname, buffer_size: cython.long = 100000):
    """Return the first parser that recognises ``fname`` as a supported format.

    Args:
        fname: Path to inspect.
        buffer_size: Buffer size handed to candidate parser constructors.

    Returns:
        Parser instance configured for the detected format.

    Raises:
        Exception: If no parser can identify the file structure.
    """
    # Note: BAMPE and BEDPE can't be automatically detected.
    ordered_parser_dict = {"BAM": BAMParser,
                           "BED": BEDParser,
                           "ELAND": ELANDResultParser,
                           "ELANDMULTI": ELANDMultiParser,
                           "ELANDEXPORT": ELANDExportParser,
                           "SAM": SAMParser,
                           "BOWTIE": BowtieParser}

    for f in ordered_parser_dict:
        p = ordered_parser_dict[f]
        t_parser = p(fname, buffer_size=buffer_size)
        debug("Testing format %s" % f)
        s = t_parser.sniff()
        if s:
            info("Detected format is: %s" % (f))
            if t_parser.is_gzipped():
                info("* Input file is gzipped.")
            return t_parser
        else:
            t_parser.close()
    raise Exception("Can't detect format!")


@cython.cfunc
def bam_fw_binary_parse(data: cython.pointer[cython.const[cython.uchar]],
                        endian_prefix=endian_prefix) -> tuple:
    """Parse a BAM SE entry.

    If in little endian system, prefix should "<", otherwise,
    ">".

    """
    thisref: cython.int
    thisstart: cython.int
    thisstrand: cython.int

    i: cython.int

    cigar_code: cython.int
    cigar_op: tuple
    n_cigar_op: cython.ushort

    bwflag: cython.ushort
    l_read_name: cython.uchar   # length for the read name

    # we skip lot of the available information in data (i.e. tag name,
    # quality etc etc) no data, return, does it really happen without
    # raising struct.error?
    if not data:
        return (-1, -1, -1)

    (n_cigar_op,  bwflag) = unpack(endian_prefix+'HH', data[12:16])

    # we filter out unmapped sequence or bad sequence or secondary or
    # supplementary alignment. we filter out 2nd mate, not a proper
    # pair, mate is unmapped
    if (bwflag & 2820) or (bwflag & 1 and (bwflag & 136 or not bwflag & 2)):
        return (-1, -1, -1)

    (thisref, thisstart) = unpack(endian_prefix+'ii', data[0:8])

    # In case of paired-end we have now skipped all possible "bad"
    # pairs in case of proper pair we have skipped the rightmost
    # one... if the leftmost pair comes we can treat it as a single
    # read, so just check the strand and calculate its start
    # position... hope I'm right!
    if bwflag & 16:
        # read mapped to minus strand; then we have to compute cigar
        # to find the rightmost position

        (l_read_name, MAPQ) = unpack(endian_prefix+'BB', data[8:10])

        # need to decipher CIGAR string
        i = 32 + l_read_name
        cigar_op = unpack(endian_prefix+'%dI' % (n_cigar_op),
                          data[i: i + n_cigar_op*4])

        for cigar_code in cigar_op:
            if cigar_code & 15 in [0, 2, 3, 7, 8]:
                # they are CIGAR op M/D/N/=/X
                thisstart += cigar_code >> 4
        thisstrand = 1
    else:
        thisstrand = 0

    return (thisref, thisstart, thisstrand)


@cython.cfunc
def bampe_pe_binary_parse(data: cython.pointer[cython.const[cython.uchar]],
                          endian_prefix=endian_prefix) -> tuple:
    """Parse a BAMPE record.

    If in little endian system, prefix should "<", otherwise,
    ">".
    """
    thisref: cython.int
    thisstart: cython.int
    thistlen: cython.int
    nextpos: cython.int
    pos: cython.int
    bwflag: cython.ushort

    # we skip lot of the available information in data (i.e. tag name,
    # quality etc etc)
    if not data:
        return (-1, -1, -1)

    bwflag = unpack(endian_prefix+'H', data[14:16])[0]

    # we filter out unmapped, bad sequence, secondary/supplementary
    # alignment we filter out other mate of paired reads, not a proper
    # pair, or mate is unmapped
    if (bwflag & 2820) or (bwflag & 1 and (bwflag & 136 or not bwflag & 2)):
        return (-1, -1, -1)

    (thisref, pos) = unpack(endian_prefix+'ii', data[0:8])

    (nextpos, thistlen) = unpack(endian_prefix+'ii', data[24:32])

    # we keep only the leftmost position which means this must be at +
    # strand. So we don't need to decipher CIGAR string.
    thisstart = min(pos, nextpos)
    # Actually, if the value unpacked is negative, then nextpos is the
    # leftmost position.
    thistlen = abs(thistlen)

    return (thisref, thisstart, thistlen)

# ------------------------------------
# Classes
# ------------------------------------


class StrandFormatError(BaseException):
    """Exception raised when strand annotations cannot be interpreted."""
    def __init__(self, string, strand):
        """Capture the offending line and strand token."""
        self.strand = strand
        self.string = string

    def __str__(self):
        """Return a descriptive error message for the malformed strand."""
        return repr("Strand information can not be recognized in this line: \"%s\",\"%s\"" %
                    (self.string, self.strand))


@cython.cclass
class GenericParser:
    """Base parser with helpers for streaming alignment-like text files.

    Attributes:
        filename: Path to the input file.
        gzipped: Whether the input stream is gzipped. (bool)
        tag_size: tag size.
        fhd: Open file handle for the input stream.
        buffer_size: Buffer size for streaming reads.
    """
    filename: str
    gzipped: bool
    tag_size: cython.int
    fhd: object
    buffer_size: cython.long

    def __init__(self, filename: str, buffer_size: cython.long = 100000):
        """Prepare the parser and open the target file.

        Args:
            filename: Path to the input alignment file.
            buffer_size: Chunk size used when reading the stream.
        """
        self.filename = filename
        self.gzipped = True
        self.tag_size = -1
        self.buffer_size = buffer_size
        # try gzip first
        f = gzip.open(filename)
        try:
            f.read(10)
        except IOError:
            # not a gzipped file
            self.gzipped = False
        f.close()
        if self.gzipped:
            # open with gzip.open, then wrap it with BufferedReader!
            # buffersize set to 10M by default.
            self.fhd = io.BufferedReader(gzip.open(filename, mode='rb'),
                                         buffer_size=READ_BUFFER_SIZE)
        else:
            # binary mode! I don't expect unicode here!
            self.fhd = io.open(filename, mode='rb')
        self.skip_first_commentlines()

    @cython.cfunc
    def skip_first_commentlines(self):
        """Advance the stream past any leading comment or header lines."""
        return

    @cython.ccall
    def tsize(self) -> cython.int:
        """Estimate tag length from a sample of valid alignments."""
        s: cython.int = 0
        n: cython.int = 0  # number of successful/valid read alignments
        m: cython.int = 0  # number of trials
        this_taglength: cython.int
        thisline: bytes

        if self.tag_size != -1:
            # if we have already calculated tag size (!= -1),  return it.
            return self.tag_size

        # try 10k times or retrieve 10 successfule alignments
        while n < 10 and m < 10000:
            m += 1
            thisline = self.fhd.readline()
            this_taglength = self.tlen_parse_line(thisline)
            if this_taglength > 0:
                # this_taglength == 0 means this line doesn't contain
                # successful alignment.
                s += this_taglength
                n += 1
        # done
        self.fhd.seek(0)
        self.skip_first_commentlines()
        if n != 0:              # else tsize = -1
            self.tag_size = cython.cast(cython.int, (s/n))
        return self.tag_size

    @cython.cfunc
    def tlen_parse_line(self, thisline: bytes) -> cython.int:
        """Return the inferred tag length for ``thisline`` or ``0`` if invalid."""
        raise NotImplementedError

    @cython.ccall
    def build_fwtrack(self):
        """Create a new ``FWTrack`` populated from the underlying stream."""
        i: cython.long
        fpos: cython.long
        strand: cython.long
        chromosome: bytes
        tmp: bytes = b""

        fwtrack = FWTrack(buffer_size=self.buffer_size)
        i = 0
        while True:
            # for each block of input
            tmp += self.fhd.read(READ_BUFFER_SIZE)
            if not tmp:
                break
            lines = tmp.split(b"\n")
            tmp = lines[-1]
            for thisline in lines[:-1]:
                (chromosome, fpos, strand) = self.fw_parse_line(thisline)
                if fpos < 0 or not chromosome:
                    # normally fw_parse_line will return -1 if the line
                    # contains no successful alignment.
                    continue
                i += 1
                if i % 1000000 == 0:
                    info(" %d reads parsed" % i)
                fwtrack.add_loc(chromosome, fpos, strand)
        # last one
        if tmp:
            (chromosome, fpos, strand) = self.fw_parse_line(tmp)
            if fpos >= 0 and chromosome:
                i += 1
                fwtrack.add_loc(chromosome, fpos, strand)
        # close file stream.
        self.close()
        return fwtrack

    @cython.ccall
    def append_fwtrack(self, fwtrack):
        """Append parsed locations to an existing ``FWTrack``."""
        i: cython.long
        fpos: cython.long
        strand: cython.long
        chromosome: bytes
        tmp: bytes = b""

        i = 0
        while True:
            # for each block of input
            tmp += self.fhd.read(READ_BUFFER_SIZE)
            if not tmp:
                break
            lines = tmp.split(b"\n")
            tmp = lines[-1]
            for thisline in lines[:-1]:
                (chromosome, fpos, strand) = self.fw_parse_line(thisline)
                if fpos < 0 or not chromosome:
                    # normally fw_parse_line will return -1 if the line
                    # contains no successful alignment.
                    continue
                i += 1
                if i % 1000000 == 0:
                    info(" %d reads parsed" % i)
                fwtrack.add_loc(chromosome, fpos, strand)

        # last one
        if tmp:
            (chromosome, fpos, strand) = self.fw_parse_line(tmp)
            if fpos >= 0 and chromosome:
                i += 1
                fwtrack.add_loc(chromosome, fpos, strand)
        # close file stream.
        self.close()
        return fwtrack

    @cython.cfunc
    def fw_parse_line(self, thisline: bytes) -> tuple:
        """Return ``(chromosome, position, strand)`` parsed from ``thisline``."""
        chromosome: bytes = b""
        fpos: cython.int = -1
        strand: cython.int = -1
        return (chromosome, fpos, strand)

    @cython.ccall
    def sniff(self):
        """Return ``True`` when the input appears compatible with this parser."""
        t: cython.int

        t = self.tsize()
        if t <= 10 or t >= 10000:  # tsize too small or too big
            self.fhd.seek(0)
            return False
        else:
            self.fhd.seek(0)
            self.skip_first_commentlines()
            return True

    @cython.ccall
    def close(self):
        """Close the underlying file handle."""
        self.fhd.close()

    @cython.ccall
    def is_gzipped(self) -> bool:
        """Report whether the underlying input stream is gzip compressed."""
        return self.gzipped


@cython.cclass
class BEDParser(GenericParser):
    """Parser for standard BED records with optional strand column."""

    @cython.cfunc
    def skip_first_commentlines(self):
        """Skip ``track``/``browser``/``#`` lines at the top of BED files."""
        l_line: cython.int
        thisline: bytes

        for thisline in self.fhd:
            l_line = len(thisline)
            if thisline and (thisline[:5] != b"track") \
               and (thisline[:7] != b"browser") \
               and (thisline[0] != 35):  # 35 is b"#"
                break

        # rewind from SEEK_CUR
        self.fhd.seek(-l_line, 1)
        return

    @cython.cfunc
    def tlen_parse_line(self, thisline: bytes) -> cython.int:
        """Return fragment length encoded in a BED line or ``0`` if invalid."""
        thisline = thisline.rstrip()
        if not thisline:
            return 0

        thisfields = thisline.split(b'\t')
        return atoi(thisfields[2]) - atoi(thisfields[1])

    @cython.cfunc
    def fw_parse_line(self, thisline: bytes) -> tuple:
        """Parse a BED entry into ``(chromosome, position, strand)``.

        Args:
            thisline: Raw line from the BED file.

        Returns:
            Tuple containing the chromosome name, 5' coordinate for the
            strand, and strand flag (0 for ``+``, 1 for ``-``).
        """
        # cdef list thisfields
        chromname: bytes
        thisfields: list

        thisline = thisline.rstrip()
        thisfields = thisline.split(b'\t')
        chromname = thisfields[0]
        try:
            if thisfields[5] == b"+":
                return (chromname,
                        atoi(thisfields[1]),
                        0)
            elif thisfields[5] == b"-":
                return (chromname,
                        atoi(thisfields[2]),
                        1)
            else:
                raise StrandFormatError(thisline, thisfields[5])
        except IndexError:
            # default pos strand if no strand
            # info can be found
            return (chromname,
                    atoi(thisfields[1]),
                    0)


@cython.cclass
class BEDPEParser(GenericParser):
    """Parser for three-column BEDPE-style fragments (chrom, left, right)."""
    n = cython.declare(cython.int, visibility='public')
    d = cython.declare(cython.float, visibility='public')

    @cython.cfunc
    def skip_first_commentlines(self):
        """Skip ``track``/``browser``/``#`` lines at the top of BEDPE files."""
        l_line: cython.int
        thisline: bytes

        for thisline in self.fhd:
            l_line = len(thisline)
            if thisline and (thisline[:5] != b"track") \
               and (thisline[:7] != b"browser") \
               and (thisline[0] != 35):  # 35 is b"#"
                break

        # rewind from SEEK_CUR
        self.fhd.seek(-l_line, 1)
        return

    @cython.cfunc
    def pe_parse_line(self, thisline: bytes):
        """Parse a fragment line into ``(chrom, left, right)`` integers."""
        thisfields: list

        thisline = thisline.rstrip()

        # still only support tabular as delimiter.
        thisfields = thisline.split(b'\t')
        try:
            return (thisfields[0],
                    atoi(thisfields[1]),
                    atoi(thisfields[2]))
        except IndexError:
            raise Exception("Less than 3 columns found at this line: %s\n" %
                            thisline)

    @cython.ccall
    def build_petrack(self):
        """Return a ``PETrackI`` constructed from the entire stream.

        Returns:
            PETrackI: Paired-end track populated from the input stream.

        Examples:
            .. code-block:: python

                from MACS3.IO.Parser import BEDPEParser
                parser = BEDPEParser("fragments.bedpe")
                petrack = parser.build_petrack()
        """
        chromosome: bytes
        left_pos: cython.int
        right_pos: cython.int
        i: cython.long = 0          # number of fragments
        m: cython.long = 0          # sum of fragment lengths
        tmp: bytes = b""

        petrack = PETrackI(buffer_size=self.buffer_size)
        add_loc = petrack.add_loc

        while True:
            # for each block of input
            tmp += self.fhd.read(READ_BUFFER_SIZE)
            if not tmp:
                break
            lines = tmp.split(b"\n")
            tmp = lines[-1]
            for thisline in lines[:-1]:
                (chromosome, left_pos, right_pos) = self.pe_parse_line(thisline)
                if left_pos < 0 or not chromosome:
                    continue
                assert right_pos > left_pos, "Right position must be larger than left position, check your BED file at line: %s" % thisline
                m += right_pos - left_pos
                i += 1
                if i % 1000000 == 0:
                    info(" %d fragments parsed" % i)
                add_loc(chromosome, left_pos, right_pos)
        # last one
        if tmp:
            (chromosome, left_pos, right_pos) = self.pe_parse_line(thisline)
            if left_pos >= 0 and chromosome:
                assert right_pos > left_pos, "Right position must be larger than left position, check your BED file at line: %s" % thisline
                i += 1
                m += right_pos - left_pos
                add_loc(chromosome, left_pos, right_pos)

        self.d = cython.cast(cython.float, m) / i
        self.n = i
        assert self.d >= 0, "Something went wrong (mean fragment size was negative)"

        self.close()
        petrack.set_rlengths({"DUMMYCHROM": 0})
        return petrack

    @cython.ccall
    def append_petrack(self, petrack):
        """Append fragments from the stream to an existing ``PETrackI``."""
        chromosome: bytes
        left_pos: cython.int
        right_pos: cython.int
        i: cython.long = 0          # number of fragments
        m: cython.long = 0          # sum of fragment lengths
        tmp: bytes = b""

        add_loc = petrack.add_loc
        while True:
            # for each block of input
            tmp += self.fhd.read(READ_BUFFER_SIZE)
            if not tmp:
                break
            lines = tmp.split(b"\n")
            tmp = lines[-1]
            for thisline in lines[:-1]:
                (chromosome, left_pos, right_pos) = self.pe_parse_line(thisline)
                if left_pos < 0 or not chromosome:
                    continue
                assert right_pos > left_pos, "Right position must be larger than left position, check your BED file at line: %s" % thisline
                m += right_pos - left_pos
                i += 1
                if i % 1000000 == 0:
                    info(" %d fragments parsed" % i)
                add_loc(chromosome, left_pos, right_pos)
        # last one
        if tmp:
            (chromosome, left_pos, right_pos) = self.pe_parse_line(thisline)
            if left_pos >= 0 and chromosome:
                assert right_pos > left_pos, "Right position must be larger than left position, check your BED file at line: %s" % thisline
                i += 1
                m += right_pos - left_pos
                add_loc(chromosome, left_pos, right_pos)

        self.d = (self.d * self.n + m) / (self.n + i)
        self.n += i

        assert self.d >= 0, "Something went wrong (mean fragment size was negative)"
        self.close()
        petrack.set_rlengths({"DUMMYCHROM": 0})
        return petrack


@cython.cclass
class ELANDResultParser(GenericParser):
    """Parser for single-end ELAND result tables."""

    @cython.cfunc
    def skip_first_commentlines(self):
        """Skip lines beginning with ``#`` before data rows."""
        l_line: cython.int
        thisline: bytes

        for thisline in self.fhd:
            l_line = len(thisline)
            if thisline and thisline[0] != 35:  # 35 is b"#"
                break

        # rewind from SEEK_CUR
        self.fhd.seek(-l_line, 1)
        return

    @cython.cfunc
    def tlen_parse_line(self, thisline: bytes) -> cython.int:
        """Return tag length for ELAND single-end entries or ``0`` if invalid."""
        thisfields: list

        thisline = thisline.rstrip()
        if not thisline:
            return 0
        thisfields = thisline.split(b'\t')
        if thisfields[1].isdigit():
            return 0
        else:
            return len(thisfields[1])

    @cython.cfunc
    def fw_parse_line(self, thisline: bytes) -> tuple:
        """Parse an ELAND result line into location and strand tuple.

        Args:
            thisline: Raw ELAND text line.

        Returns:
            ``(chromosome, position, strand)`` with ``position`` set to
            ``-1`` when the line does not encode a valid alignment.
        """
        chromname: bytes
        strand: bytes
        thistaglength: cython.int
        thisfields: list

        thisline = thisline.rstrip()
        if not thisline:
            return (b"", -1, -1)

        thisfields = thisline.split(b'\t')
        thistaglength = len(thisfields[1])

        if len(thisfields) <= 6:
            return (b"", -1, -1)

        try:
            chromname = thisfields[6]
            chromname = chromname[:chromname.rindex(b".fa")]
        except ValueError:
            pass

        if thisfields[2] == b"U0" or thisfields[2] == b"U1" or thisfields[2] == b"U2":
            # allow up to 2 mismatches...
            strand = thisfields[8]
            if strand == b"F":
                return (chromname,
                        atoi(thisfields[7]) - 1,
                        0)
            elif strand == b"R":
                return (chromname,
                        atoi(thisfields[7]) + thistaglength - 1,
                        1)
            else:
                raise StrandFormatError(thisline, strand)
        else:
            return (b"", -1, -1)


@cython.cclass
class ELANDMultiParser(GenericParser):
    """Parser for ELAND multi-hit reports (``s_N_eland_multi`` format)."""

    @cython.cfunc
    def skip_first_commentlines(self):
        """Skip lines beginning with ``#`` before data rows."""
        l_line: cython.int
        thisline: bytes

        for thisline in self.fhd:
            l_line = len(thisline)
            if thisline and thisline[0] != 35:  # 35 is b"#"
                break

        # rewind from SEEK_CUR
        self.fhd.seek(-l_line, 1)
        return

    @cython.cfunc
    def tlen_parse_line(self, thisline: bytes) -> cython.int:
        """Return tag length for ELAND multi entries or ``0`` if ambiguous."""
        thisline = thisline.rstrip()
        if not thisline:
            return 0
        thisfields = thisline.split(b'\t')
        if thisfields[1].isdigit():
            return 0
        else:
            return len(thisfields[1])

    @cython.cfunc
    def fw_parse_line(self, thisline: bytes) -> tuple:
        """Parse ELAND multi-format line into a single-hit location tuple.

        Args:
            thisline: Raw ELAND multi line.

        Returns:
            ``(chromosome, position, strand)`` or ``(b"", -1, -1)`` if the
            entry is not uniquely mappable.
        """
        # thistagname: bytes
        pos: bytes
        strand: bytes
        thisfields: list
        thistaglength: cython.int
        thistaghits: cython.int

        if not thisline:
            return (b"", -1, -1)
        thisline = thisline.rstrip()
        if not thisline:
            return (b"", -1, -1)

        thisfields = thisline.split(b'\t')
        # thistagname = thisfields[0]        # name of tag
        thistaglength = len(thisfields[1])  # length of tag

        if len(thisfields) < 4:
            return (b"", -1, -1)
        else:
            thistaghits = sum([cython.cast(cython.int, x) for x in thisfields[2].split(b':')])
            if thistaghits > 1:
                # multiple hits
                return (b"", -1, -1)
            else:
                (chromname, pos) = thisfields[3].split(b':')

                try:
                    chromname = chromname[:chromname.rindex(b".fa")]
                except ValueError:
                    pass

                strand = pos[-2]
                if strand == b"F":
                    return (chromname,
                            cython.cast(cython.int, pos[:-2]) - 1,
                            0)
                elif strand == b"R":
                    return (chromname,
                            cython.cast(cython.int, pos[:-2]) + thistaglength - 1,
                            1)
                else:
                    raise StrandFormatError(thisline, strand)


@cython.cclass
class ELANDExportParser(GenericParser):
    """Parser for ELAND export tab-delimited files."""
    @cython.cfunc
    def skip_first_commentlines(self):
        """Skip lines beginning with ``#`` before data rows."""
        l_line: cython.int
        thisline: bytes

        for thisline in self.fhd:
            l_line = len(thisline)
            if thisline and thisline[0] != 35:  # 35 is b"#"
                break

        # rewind from SEEK_CUR
        self.fhd.seek(-l_line, 1)
        return

    @cython.cfunc
    def tlen_parse_line(self, thisline: bytes) -> cython.int:
        """Return tag length for ELAND export entries or ``0`` if invalid."""
        thisline = thisline.rstrip()
        if not thisline:
            return 0
        thisfields = thisline.split(b'\t')
        if len(thisfields) > 12 and thisfields[12]:
            # a successful alignment has over 12 columns
            return len(thisfields[8])
        else:
            return 0

    @cython.cfunc
    def fw_parse_line(self, thisline: bytes) -> tuple:
        """Parse an ELAND export entry to ``(chromosome, position, strand)``.

        Args:
            thisline: Raw ELAND export line.

        Returns:
            Location tuple when the line contains a successful alignment;
            otherwise ``(b"", -1, -1)``.
        """
        # thisname: bytes
        strand: bytes
        thistaglength: cython.int
        thisfields: list

        thisline = thisline.rstrip()
        if not thisline:
            return (b"", -1, -1)

        thisfields = thisline.split(b"\t")

        if len(thisfields) > 12 and thisfields[12]:
            # thisname = b":".join(thisfields[0:6])
            thistaglength = len(thisfields[8])
            strand = thisfields[13]
            if strand == b"F":
                return (thisfields[10], atoi(thisfields[12]) - 1, 0)
            elif strand == b"R":
                return (thisfields[10], atoi(thisfields[12]) + thistaglength - 1, 1)
            else:
                raise StrandFormatError(thisline, strand)
        else:
            return (b"", -1, -1)


# Contributed by Davide, modified by Tao

@cython.cclass
class SAMParser(GenericParser):
    """Parser for SAM alignment text files with standard SAM flags."""

    @cython.cfunc
    def skip_first_commentlines(self):
        """Skip SAM header lines beginning with ``@``."""
        l_line: cython.int
        thisline: bytes

        for thisline in self.fhd:
            l_line = len(thisline)
            if thisline and thisline[0] != 64:  # 64 is b"@"
                break

        # rewind from SEEK_CUR
        self.fhd.seek(-l_line, 1)
        return

    @cython.cfunc
    def tlen_parse_line(self, thisline: bytes) -> cython.int:
        """Return read length for valid SAM records or ``0`` if filtered."""
        thisfields: list
        bwflag: cython.int

        thisline = thisline.rstrip()
        if not thisline:
            return 0
        thisfields = thisline.split(b'\t')
        bwflag = atoi(thisfields[1])
        if bwflag & 4 or bwflag & 512 or bwflag & 256 or bwflag & 2048:
            # unmapped sequence or bad sequence or 2nd or sup alignment
            return 0

        if bwflag & 1:
            # paired read. We should only keep sequence if the mate is mapped
            # and if this is the left mate, all is within  the flag!
            if not bwflag & 2:
                return 0   # not a proper pair
            if bwflag & 8:
                return 0   # the mate is unmapped
            # From Benjamin Schiller https://github.com/benjschiller
            if bwflag & 128:
                # this is not the first read in a pair
                return 0
        return len(thisfields[9])

    @cython.cfunc
    def fw_parse_line(self, thisline: bytes) -> tuple:
        """Parse a SAM alignment into ``(chromosome, position, strand)``.

        Args:
            thisline: Raw SAM alignment line.

        Returns:
            Tuple identifying the leftmost coordinate for forward strands or
            the rightmost coordinate for reverse strands. Returns
            ``(b"", -1, -1)`` for reads that do not satisfy mapping filters.
        """
        # thistagname: bytes
        thisref: bytes
        thisfields: list
        bwflag: cython.int
        thisstrand: cython.int
        thisstart: cython.int

        thisline = thisline.rstrip()
        if not thisline:
            return (b"", -1, -1)
        thisfields = thisline.split(b'\t')
        # thistagname = thisfields[0]         # name of tag
        thisref = thisfields[2]
        bwflag = atoi(thisfields[1])
        CIGAR = thisfields[5]

        if (bwflag & 2820) or (bwflag & 1 and (bwflag & 136 or not bwflag & 2)):
            return (b"", -1, -1)

        # if bwflag & 4 or bwflag & 512 or bwflag & 256 or bwflag & 2048:
        #    return (b"", -1, -1)       #unmapped sequence or bad sequence or 2nd or sup alignment
        # if bwflag & 1:
        #    # paired read. We should only keep sequence if the mate is mapped
        #    # and if this is the left mate, all is within  the flag!
        #    if not bwflag & 2:
        #        return (b"", -1, -1)   # not a proper pair
        #    if bwflag & 8:
        #        return (b"", -1, -1)   # the mate is unmapped
        #    # From Benjamin Schiller https://github.com/benjschiller
        #    if bwflag & 128:
        #        # this is not the first read in a pair
        #        return (b"", -1, -1)
        #    # end of the patch
        # In case of paired-end we have now skipped all possible "bad" pairs
        # in case of proper pair we have skipped the rightmost one... if the leftmost pair comes
        # we can treat it as a single read, so just check the strand and calculate its
        # start position... hope I'm right!
        if bwflag & 16:
            # minus strand, we have to decipher CIGAR string
            thisstrand = 1
            thisstart = atoi(thisfields[3]) - 1 + sum([cython.cast(cython.int, x) for x in findall(b"(\\d+)[MDNX=]", CIGAR)])  #reverse strand should be shifted alen bp
        else:
            thisstrand = 0
            thisstart = atoi(thisfields[3]) - 1

        try:
            thisref = thisref[:thisref.rindex(b".fa")]
        except ValueError:
            pass

        return (thisref, thisstart, thisstrand)


@cython.cclass
class BAMParser(GenericParser):
    """Parser for BAM binary alignment files."""
    def __init__(self, filename: str,
                 buffer_size: cython.long = 100000):
        """Prepare the parser and open the BAM file.

        Args:
            filename: Path to the BAM file.
            buffer_size: Chunk size used when reading BGZF blocks.
        """
        self.filename = filename
        self.gzipped = True
        self.tag_size = -1
        self.buffer_size = buffer_size
        # try gzip first
        f = gzip.open(filename)
        try:
            f.read(10)
        except IOError:
            # not a gzipped file
            self.gzipped = False
        f.close()
        if self.gzipped:
            # open with gzip.open, then wrap it with BufferedReader!
            self.fhd = io.BufferedReader(gzip.open(filename, mode='rb'),
                                         buffer_size=READ_BUFFER_SIZE)
        else:
            # binary mode! I don't expect unicode here!
            self.fhd = io.open(filename, mode='rb')

    @cython.ccall
    def sniff(self):
        """Return ``True`` if the file begins with the BAM magic string."""
        magic_header: bytes
        tsize: cython.int

        magic_header = self.fhd.read(3)
        if magic_header == b"BAM":
            tsize = self.tsize()
            if tsize > 0:
                self.fhd.seek(0)
                return True
            else:
                self.fhd.seek(0)
                raise Exception("File is not of a valid BAM format! %d" %
                                tsize)
        else:
            self.fhd.seek(0)
            return False

    @cython.ccall
    def tsize(self) -> cython.int:
        """Get tag size from BAM file -- read l_seq field.

        Refer to: http://samtools.sourceforge.net/SAM1.pdf

        * This may not work for BAM file from bedToBAM (bedtools),
        since the l_seq field seems to be 0.
        """
        x: cython.int
        header_len: cython.int
        nc: cython.int
        nlength: cython.int
        n: cython.int = 0          # successful read of tag size
        s: cython.double = 0        # sum of tag sizes

        if self.tag_size != -1:
            # if we have already calculated tag size (!= -1),  return it.
            return self.tag_size

        fseek = self.fhd.seek
        fread = self.fhd.read
        ftell = self.fhd.tell
        # move to pos 4, there starts something
        fseek(4)
        header_len = unpack('<i', fread(4))[0]
        fseek(header_len + ftell())
        # get the number of chromosome
        nc = unpack('<i', fread(4))[0]
        for x in range(nc):
            # read each chromosome name
            nlength = unpack('<i', fread(4))[0]
            # jump over chromosome size, we don't need it
            fread(nlength)
            fseek(ftell() + 4)
        while n < 10:
            entrylength = unpack('<i', fread(4))[0]
            data = fread(entrylength)
            a = unpack('<i', data[16:20])[0]
            s += a
            n += 1
        fseek(0)
        self.tag_size = cython.cast(cython.int, (s/n))
        return self.tag_size

    @cython.ccall
    def get_references(self) -> tuple:
        """Return ``(references, lengths)`` extracted from the BAM header."""
        header_len: cython.int
        x: cython.int
        nc: cython.int
        nlength: cython.int
        refname: bytes
        references: list = []
        rlengths: dict = {}

        fseek = self.fhd.seek
        fread = self.fhd.read
        ftell = self.fhd.tell
        # move to pos 4, there starts something
        fseek(4)
        header_len = unpack('<i', fread(4))[0]
        fseek(header_len + ftell())
        # get the number of chromosome
        nc = unpack('<i', fread(4))[0]
        for x in range(nc):
            # read each chromosome name
            nlength = unpack('<i', fread(4))[0]
            refname = fread(nlength)[:-1]
            references.append(refname)
            # don't jump over chromosome size
            # we can use it to avoid falling of chrom ends during peak calling
            rlengths[refname] = unpack('<i', fread(4))[0]
        return (references, rlengths)

    @cython.ccall
    def build_fwtrack(self):
        """Append uniquely mapped reads to an existing ``FWTrack``."""
        i: cython.long = 0          # number of reads kept
        entrylength: cython.int
        fpos: cython.int
        strand: cython.int
        chrid: cython.int
        references: list
        rlengths: dict

        fwtrack = FWTrack(buffer_size=self.buffer_size)
        # after this, ptr at list of alignments
        references, rlengths = self.get_references()  
        # fseek = self.fhd.seek
        fread = self.fhd.read
        # ftell = self.fhd.tell

        while True:
            try:
                entrylength = unpack("<i", fread(4))[0]
            except struct.error:
                break
            (chrid, fpos, strand) = bam_fw_binary_parse(fread(entrylength))
            if chrid == -1:
                continue
            fwtrack.add_loc(references[chrid], fpos, strand)
            i += 1
            if i % 1000000 == 0:
                info(" %d reads parsed" % i)

        # print(f"{references[chrid]:},{fpos:},{strand:}")
        info("%d reads have been read." % i)
        self.fhd.close()
        fwtrack.set_rlengths(rlengths)
        return fwtrack

    @cython.ccall
    def append_fwtrack(self, fwtrack):
        """Append uniquely mapped reads to an existing ``FWTrack``."""
        i: cython.long = 0          # number of reads kept
        entrylength: cython.int
        fpos: cython.int
        strand: cython.int
        chrid: cython.int
        references: list
        rlengths: dict

        references, rlengths = self.get_references()
        # fseek = self.fhd.seek
        fread = self.fhd.read
        # ftell = self.fhd.tell

        while True:
            try:
                entrylength = unpack('<i', fread(4))[0]
            except struct.error:
                break
            (chrid, fpos, strand) = bam_fw_binary_parse(fread(entrylength))
            if chrid == -1:
                continue
            fwtrack.add_loc(references[chrid], fpos, strand)
            i += 1
            if i % 1000000 == 0:
                info(" %d reads parsed" % i)

        info("%d reads have been read." % i)
        self.fhd.close()
        # fwtrack.finalize()
        # this is the problematic part. If fwtrack is finalized, then
        # it's impossible to increase the length of it in a step of
        # buffer_size for multiple input files.
        fwtrack.set_rlengths(rlengths)
        return fwtrack


@cython.cclass
class BAMPEParser(BAMParser):
    """Parser for paired-end BAM files that yields fragment spans.

    Attributes:
        n: Total number of fragments.
        d: Average fragment length.
        
    """
    # total number of fragments
    n = cython.declare(cython.int, visibility='public')
    # the average length of fragments
    d = cython.declare(cython.float, visibility='public')

    @cython.ccall
    def build_petrack(self):
        """Return a ``PETrackI`` populated with inferred fragments.

        Returns:
            PETrackI: Paired-end track populated from BAM pairs.

        Examples:
            .. code-block:: python

                from MACS3.IO.Parser import BAMPEParser
                parser = BAMPEParser("reads.bam")
                petrack = parser.build_petrack()
        """
        i: cython.long = 0          # number of fragments kept
        m: cython.long = 0          # sum of fragment lengths
        entrylength: cython.int
        fpos: cython.int
        chrid: cython.int
        tlen: cython.int
        references: list
        rlengths: dict

        petrack = PETrackI(buffer_size=self.buffer_size)
        references, rlengths = self.get_references()
        # fseek = self.fhd.seek
        fread = self.fhd.read
        # ftell = self.fhd.tell

        # for convenience, only count valid pairs
        add_loc = petrack.add_loc
        err = struct.error

        while True:
            try:
                entrylength = unpack('<i', fread(4))[0]
            except err:
                # e1 += 1
                break
            (chrid, fpos, tlen) = bampe_pe_binary_parse(fread(entrylength))
            if chrid == -1:
                # e2 += 1
                continue
            add_loc(references[chrid], fpos, fpos + tlen)
            m += tlen
            i += 1
            if i % 1000000 == 0:
                info(" %d fragments parsed" % i)

        info("%d fragments have been read." % i)
        self.d = m / i
        self.n = i
        self.fhd.close()
        petrack.set_rlengths(rlengths)
        return petrack

    @cython.ccall
    def append_petrack(self, petrack):
        """Append inferred fragments to an existing ``PETrackI``.

        Args:
            petrack: Existing paired-end track to append to.

        Returns:
            PETrackI: The updated track instance.

        Examples:
            .. code-block:: python

                from MACS3.IO.Parser import BAMPEParser
                parser = BAMPEParser("reads.bam")
                petrack = parser.build_petrack()
                # Later, append more fragments from another file:
                parser2 = BAMPEParser("more_reads.bam")
                petrack = parser2.append_petrack(petrack)
        """
        i: cython.long = 0          # number of fragments
        m: cython.long = 0          # sum of fragment lengths
        entrylength: cython.int
        fpos: cython.int
        chrid: cython.int
        tlen: cython.int
        references: list
        rlengths: dict

        references, rlengths = self.get_references()
        fread = self.fhd.read

        # for convenience, only count valid pairs
        add_loc = petrack.add_loc
        err = struct.error

        while True:
            try:
                entrylength = unpack('<i', fread(4))[0]
            except err:
                break
            (chrid, fpos, tlen) = bampe_pe_binary_parse(fread(entrylength))
            if chrid == -1:
                continue
            add_loc(references[chrid], fpos, fpos + tlen)
            m += tlen
            i += 1
            if i % 1000000 == 0:
                info(" %d fragments parsed" % i)

        info("%d fragments have been read." % i)
        self.d = (self.d * self.n + m) / (self.n + i)
        self.n += i
        self.fhd.close()
        petrack.set_rlengths(rlengths)
        return petrack


@cython.cclass
class BowtieParser(GenericParser):
    """Parser for Bowtie or Maqview single-end map files."""
    @cython.cfunc
    def tlen_parse_line(self, thisline: bytes) -> cython.int:
        """Return read length for Bowtie map entries or ``0`` if invalid."""
        thisfields: list

        thisline = thisline.rstrip()
        if not thisline:
            return (b"", -1, -1)
        if thisline[0] == b"#":
            return (b"", -1, -1)  # comment line is skipped
        thisfields = thisline.split(b'\t')  # I hope it will never bring me more trouble
        return len(thisfields[4])

    @cython.cfunc
    def fw_parse_line(self, thisline: bytes) -> tuple:
        """
        The following definition comes from bowtie website:

        The bowtie aligner outputs each alignment on a separate
        line. Each line is a collection of 8 fields separated by tabs;
        from left to right, the fields are:

        1. Name of read that aligned

        2. Orientation of read in the alignment, - for reverse
        complement, + otherwise

        3. Name of reference sequence where alignment occurs, or
        ordinal ID if no name was provided

        4. 0-based offset into the forward reference strand where
        leftmost character of the alignment occurs

        5. Read sequence (reverse-complemented if orientation is -)

        6. ASCII-encoded read qualities (reversed if orientation is
        -). The encoded quality values are on the Phred scale and the
        encoding is ASCII-offset by 33 (ASCII char !).

        7. Number of other instances where the same read aligns
        against the same reference characters as were aligned against
        in this alignment. This is not the number of other places the
        read aligns with the same number of mismatches. The number in
        this column is generally not a good proxy for that number
        (e.g., the number in this column may be '0' while the number
        of other alignments with the same number of mismatches might
        be large). This column was previously described as 'Reserved'.

        8. Comma-separated list of mismatch descriptors. If there are
        no mismatches in the alignment, this field is empty. A single
        descriptor has the format offset:reference-base>read-base. The
        offset is expressed as a 0-based offset from the high-quality
        (5') end of the read.

        """
        thisfields: list
        chromname: bytes

        thisline = thisline.rstrip()
        if not thisline:
            return (b"", -1, -1)
        if thisline[0] == b"#":
            return (b"", -1, -1)  # comment line is skipped
        # I hope it will never bring me more trouble
        thisfields = thisline.split(b'\t')

        chromname = thisfields[2]
        try:
            chromname = chromname[:chromname.rindex(b".fa")]
        except ValueError:
            pass

            if thisfields[1] == b"+":
                return (chromname,
                        atoi(thisfields[3]),
                        0)
            elif thisfields[1] == b"-":
                return (chromname,
                        atoi(thisfields[3]) + len(thisfields[4]),
                        1)
            else:
                raise StrandFormatError(thisline, thisfields[1])


@cython.cclass
class FragParser(GenericParser):
    """Parser for scATAC fragment TSV files with barcode counts."""
    n = cython.declare(cython.int, visibility='public')
    d = cython.declare(cython.float, visibility='public')

    @cython.cfunc
    def skip_first_commentlines(self):
        """Skip ``track``/``browser``/``#`` lines at the top of fragment files."""
        l_line: cython.int
        thisline: bytes

        for thisline in self.fhd:
            l_line = len(thisline)
            if thisline and (thisline[:5] != b"track") \
               and (thisline[:7] != b"browser") \
               and (thisline[0] != 35):  # 35 is b"#"
                break

        # rewind from SEEK_CUR
        self.fhd.seek(-l_line, 1)
        return

    @cython.cfunc
    def pe_parse_line(self, thisline: bytes):
        """Parse a fragment line into ``(chrom, left, right, barcode, count)``."""
        thisfields: list
        thiscount: cython.ushort

        thisline = thisline.rstrip()

        # still only support tabular as delimiter.
        thisfields = thisline.split(b'\t')
        try:
            try:
                thiscount = atoi(thisfields[4])
            except OverflowError:
                thiscount = 65535
                warn(f"The count in this line is over 65535, and will be capped at 65535: {thisline}")
            return (thisfields[0],
                    atoi(thisfields[1]),
                    atoi(thisfields[2]),
                    thisfields[3],
                    thiscount)
        except IndexError:
            raise Exception("Less than 5 columns found at this line: %s\n" %
                            thisline)

    @cython.ccall
    def build_petrack(self, max_count=0):
        """Return a ``PETrackII`` populated with fragments and barcodes.

        Args:
            max_count: Optional cap applied to per-fragment count values.

        Returns:
            PETrackII: Paired-end track with barcode and count metadata.

        Examples:
            .. code-block:: python

                from MACS3.IO.Parser import FragParser
                parser = FragParser("fragments.tsv.gz")
                petrack = parser.build_petrack(max_count=10)
        """
        chromosome: bytes
        left_pos: cython.int
        right_pos: cython.int
        barcode: bytes
        count: cython.ushort
        i: cython.long = 0          # number of fragments
        m: cython.long = 0          # sum of fragment lengths
        tmp: bytes = b""

        petrack = PETrackII(buffer_size=self.buffer_size)
        add_loc = petrack.add_loc

        while True:
            # for each block of input
            tmp += self.fhd.read(READ_BUFFER_SIZE)
            if not tmp:
                break
            lines = tmp.split(b"\n")
            tmp = lines[-1]
            for thisline in lines[:-1]:
                (chromosome, left_pos, right_pos, barcode, count) = self.pe_parse_line(thisline)
                if left_pos < 0 or not chromosome:
                    continue
                assert right_pos > left_pos, "Right position must be larger than left position, check your BED file at line: %s" % thisline
                m += right_pos - left_pos
                i += 1
                if i % 1000000 == 0:
                    info(" %d fragments parsed" % i)
                if max_count:
                    count = min(count, max_count)
                add_loc(chromosome, left_pos, right_pos, barcode, count)
        # last one
        if tmp:
            (chromosome, left_pos, right_pos, barcode, count) = self.pe_parse_line(thisline)
            if left_pos >= 0 and chromosome:
                assert right_pos > left_pos, "Right position must be larger than left position, check your BED file at line: %s" % thisline
                i += 1
                m += right_pos - left_pos
                if max_count:
                    count = min(count, max_count)                
                add_loc(chromosome, left_pos, right_pos, barcode, count)

        self.d = cython.cast(cython.float, m) / i
        self.n = i
        assert self.d >= 0, "Something went wrong (mean fragment size was negative)"

        self.close()
        petrack.set_rlengths({"DUMMYCHROM": 0})
        return petrack

    @cython.ccall
    def append_petrack(self, petrack, max_count=0):
        """Append barcode-aware fragments to an existing ``PETrackI``.

        Args:
            petrack: Existing paired-end track to append to.
            max_count: Optional cap applied to per-fragment count values.

        Returns:
            PETrackII: The updated track instance.

        Examples:
            .. code-block:: python

                from MACS3.IO.Parser import FragParser
                parser = FragParser("fragments.tsv.gz")
                petrack = parser.build_petrack()
                parser2 = FragParser("more_fragments.tsv.gz")
                petrack = parser2.append_petrack(petrack, max_count=10)
        """
        chromosome: bytes
        left_pos: cython.int
        right_pos: cython.int
        barcode: bytes
        count: cython.ushort
        i: cython.long = 0          # number of fragments
        m: cython.long = 0          # sum of fragment lengths
        tmp: bytes = b""

        add_loc = petrack.add_loc
        while True:
            # for each block of input
            tmp += self.fhd.read(READ_BUFFER_SIZE)
            if not tmp:
                break
            lines = tmp.split(b"\n")
            tmp = lines[-1]
            for thisline in lines[:-1]:
                (chromosome, left_pos, right_pos, barcode, count) = self.pe_parse_line(thisline)
                if left_pos < 0 or not chromosome:
                    continue
                assert right_pos > left_pos, "Right position must be larger than left position, check your BED file at line: %s" % thisline
                m += right_pos - left_pos
                i += 1
                if i % 1000000 == 0:
                    info(" %d fragments parsed" % i)
                if max_count:
                    count = min(count, max_count)                    
                add_loc(chromosome, left_pos, right_pos, barcode, count)
        # last one
        if tmp:
            (chromosome, left_pos, right_pos, barcode, count) = self.pe_parse_line(thisline)
            if left_pos >= 0 and chromosome:
                assert right_pos > left_pos, "Right position must be larger than left position, check your BED file at line: %s" % thisline
                i += 1
                m += right_pos - left_pos
                if max_count:
                    count = min(count, max_count)                
                add_loc(chromosome, left_pos, right_pos, barcode, count)

        self.d = (self.d * self.n + m) / (self.n + i)
        self.n += i

        assert self.d >= 0, "Something went wrong (mean fragment size was negative)"
        self.close()
        petrack.set_rlengths({"DUMMYCHROM": 0})
        return petrack



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2025-11-14 19:08:47 Tao Liu>

"""Module for filter duplicate tags from paired-end data

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

# ------------------------------------
# Python modules
# ------------------------------------
import io
import sys
from array import array as pyarray
from collections import Counter,defaultdict
# ------------------------------------
# MACS3 modules
# ------------------------------------
from MACS3.Signal.Pileup import (quick_pileup,
                                 over_two_pv_array,
                                 se_all_in_one_pileup)
from MACS3.Signal.BedGraph import (bedGraphTrackI,
                                   bedGraphTrackII)
from MACS3.Signal.PileupV2 import (pileup_from_LR_hmmratac,
                                   pileup_from_LRC)
from MACS3.Signal.Region import Regions
# ------------------------------------
# Other modules
# ------------------------------------
import cython
import numpy as np
import cython.cimports.numpy as cnp
from cython.cimports.cpython import bool
from cython.cimports.libc.stdint import INT32_MAX as INT_MAX

from MACS3.Utilities.Logger import logging

logger = logging.getLogger(__name__)
debug = logger.debug
info = logger.info

# Let numpy enforce PE-ness using ndarray, gives bonus speedup when sorting
# PE data doesn't have strandedness


@cython.cclass
class PETrackI:
    """In-memory paired-end fragment container grouped by chromosome.
    
    The track exposes utilities for sorting, filtering, downsampling, and pileup
    generation on numpy structured arrays of left/right coordinates.
    """
    locations = cython.declare(dict, visibility="public")
    size = cython.declare(dict, visibility="public")
    buf_size = cython.declare(dict, visibility="public")
    is_sorted = cython.declare(bool, visibility="public")
    total = cython.declare(cython.ulong, visibility="public")
    annotation = cython.declare(str, visibility="public")
    # rlengths: reference chromosome lengths dictionary
    rlengths = cython.declare(dict, visibility="public")
    buffer_size = cython.declare(cython.long, visibility="public")
    length = cython.declare(cython.ulonglong, visibility="public")
    average_template_length = cython.declare(cython.float, visibility="public")
    is_destroyed: bool

    def __init__(self, anno: str = "", buffer_size: cython.long = 100000):
        """Initialize an empty paired-end track.
        
        Parameters
        ----------
        anno : str, optional
            Annotation label retained with the track metadata.
        buffer_size : int, optional
            Number of fragment slots allocated per growth chunk for each chromosome.
        """
        # dictionary with chrname as key, nparray with
        # [('l','i4'),('r','i4')] as value
        self.locations = {}
        # dictionary with chrname as key, size of the above nparray as value
        # size is to remember the number of the fragments added to this chromosome
        self.size = {}
        # dictionary with chrname as key, size of the above nparray as value
        self.buf_size = {}
        self.is_sorted = False
        self.total = 0           # total fragments
        self.annotation = anno   # need to be figured out
        self.rlengths = {}
        self.buffer_size = buffer_size
        self.length = 0
        self.average_template_length = 0.0
        self.is_destroyed = False

    @cython.ccall
    def add_loc(self, chromosome: bytes,
                start: cython.int, end: cython.int):
        """Append a paired-end fragment to the track.
        
        Parameters
        ----------
        chromosome : bytes
            Chromosome name (as bytes) that owns the fragment.
        start : int
            Zero-based start coordinate of the fragment (5' end).
        end : int
            Zero-based end coordinate of the fragment (3' end).
        
        Notes
        -----
        Fragments are stored in structured numpy arrays keyed by chromosome, and
        the running fragment count and total template length are updated in place.

        Examples
        --------
        .. code-block:: python

            from MACS3.Signal.PairedEndTrack import PETrackI

            pe = PETrackI()
            pe.add_loc(b"chr1", 10, 40)
            pe.add_loc(b"chr1", 50, 90)
        """
        i: cython.int

        if chromosome not in self.locations:
            self.buf_size[chromosome] = self.buffer_size
            # note: ['l'] is the leftmost end, ['r'] is the rightmost end of fragment.
            self.locations[chromosome] = np.zeros(shape=self.buffer_size,
                                                  dtype=[('l', 'i4'), ('r', 'i4')])
            self.locations[chromosome][0] = (start, end)
            self.size[chromosome] = 1
        else:
            i = self.size[chromosome]
            if self.buf_size[chromosome] == i:
                self.buf_size[chromosome] += self.buffer_size
                self.locations[chromosome].resize((self.buf_size[chromosome]),
                                                  refcheck=False)
            self.locations[chromosome][i] = (start, end)
            self.size[chromosome] = i + 1
        self.length += end - start
        return

    @cython.ccall
    def destroy(self):
        """Release numpy buffers held by the track.
        
        All per-chromosome arrays are resized to zero so the memory footprint is
        returned to the allocator, and the track is marked as destroyed.
        """
        chrs: set
        chromosome: bytes

        chrs = self.get_chr_names()
        for chromosome in sorted(chrs):
            if chromosome in self.locations:
                self.locations[chromosome].resize(self.buffer_size,
                                                  refcheck=False)
                self.locations[chromosome].resize(0,
                                                  refcheck=False)
                self.locations[chromosome] = None
                self.locations.pop(chromosome)
        self.is_destroyed = True
        return

    @cython.ccall
    def set_rlengths(self, rlengths: dict) -> bool:
        """Attach reference chromosome lengths to the track.
        
        Parameters
        ----------
        rlengths : dict
            Mapping from chromosome name (bytes) to reference length.
        
        Returns
        -------
        bool
            True when the length mapping has been updated.
        
        Notes
        -----
        Any chromosome stored in the track but missing from ``rlengths`` is
        assigned ``INT_MAX`` so downstream bounds checks can succeed.
        """
        valid_chroms: set
        missed_chroms: set
        chrom: bytes

        valid_chroms = set(self.locations.keys()).intersection(rlengths.keys())
        for chrom in sorted(valid_chroms):
            self.rlengths[chrom] = rlengths[chrom]
        missed_chroms = set(self.locations.keys()).difference(rlengths.keys())
        for chrom in sorted(missed_chroms):
            self.rlengths[chrom] = INT_MAX
        return True

    @cython.ccall
    def get_rlengths(self) -> dict:
        """Return the reference chromosome lengths associated with the track.
        
        Returns
        -------
        dict
            Mapping from chromosome name (bytes) to reference length. Chromosomes
            without a recorded length default to ``INT_MAX``.
        """
        if not self.rlengths:
            self.rlengths = dict([(k, INT_MAX) for k in self.locations.keys()])
        return self.rlengths

    @cython.ccall
    def finalize(self):
        """Shrink backing arrays and sort fragments in place.
        
        Each per-chromosome array is resized to the observed fragment count, sorted
        by the left and right coordinates, and the aggregate counters ``total`` and
        ``average_template_length`` are refreshed. Call this after loading data.
        """
        c: bytes
        chrnames: set

        self.total = 0

        chrnames = self.get_chr_names()

        for c in chrnames:
            self.locations[c].resize((self.size[c]), refcheck=False)
            if self.size[c] == 0:
                if c in self.size:
                    del self.size[c]
                if c in self.locations:
                    del self.locations[c]
                if c in self.rlengths:
                    del self.rlengths[c]
                continue
            self.locations[c].sort(order=['l', 'r'])
            self.total += self.size[c]

        self.is_sorted = True
        self.average_template_length = cython.cast(cython.float, self.length) / self.total
        return

    @cython.ccall
    def get_locations_by_chr(self, chromosome: bytes):
        """Return the fragment array for a chromosome.
        
        Parameters
        ----------
        chromosome : bytes
            Chromosome name, provided as bytes.
        
        Returns
        -------
        numpy.ndarray
            Structured array with ``('l', 'i4')`` and ``('r', 'i4')`` fields.
        
        Raises
        ------
        Exception
            If the chromosome is not present in the track.
        """
        if chromosome in self.locations:
            return self.locations[chromosome]
        else:
            raise Exception("No such chromosome name (%s) in TrackI object!\n" % (chromosome))

    @cython.ccall
    def get_chr_names(self) -> set:
        """Return the set of chromosome names stored in the track.
        
        Returns
        -------
        set
            Chromosome names (bytes) that currently have fragments.
        """
        return set(self.locations.keys())

    @cython.ccall
    def sort(self):
        """Sort fragments for each chromosome by genomic coordinate.
        
        Fragments are ordered first by their left coordinate and then by their right
        coordinate. The ``is_sorted`` flag is set to ``True`` when sorting completes.
        """
        c: bytes
        chrnames: set

        chrnames = self.get_chr_names()

        for c in chrnames:
            self.locations[c].sort(order=['l', 'r'])  # sort by the leftmost location
        self.is_sorted = True
        return

    @cython.ccall
    def count_fraglengths(self) -> dict:
        """Count observed fragment lengths across the track.
        
        Returns
        -------
        dict
            Mapping from fragment length to observed count, useful for downstream
            models such as HMMRATAC.
        """
        sizes: cnp.ndarray(cnp.int32_t, ndim=1)
        s: cython.int
        locs: cnp.ndarray
        chrnames: list
        i: cython.int

        counter = Counter()
        chrnames = list(self.get_chr_names())
        for i in range(len(chrnames)):
            locs = self.locations[chrnames[i]]
            sizes = locs['r'] - locs['l']
            for s in sizes:
                counter[s] += 1
        return dict(counter)

    @cython.ccall
    def fraglengths(self) -> cnp.ndarray:
        """Return all fragment lengths as a single array.
        
        Returns
        -------
        numpy.ndarray
            Concatenated array of ``end - start`` for every stored fragment across
            chromosomes.
        """
        sizes: cnp.ndarray(np.int32_t, ndim=1)
        locs: cnp.ndarray
        chrnames: list
        i: cython.int

        chrnames = list(self.get_chr_names())
        locs = self.locations[chrnames[0]]
        sizes = locs['r'] - locs['l']
        for i in range(1, len(chrnames)):
            locs = self.locations[chrnames[i]]
            sizes = np.concatenate((sizes, locs['r'] - locs['l']))
        return sizes

    @cython.boundscheck(False)  # do not check that np indices are valid
    @cython.ccall
    def exclude(self, regions):
        """Remove fragments that overlap the provided exclusion regions.
        
        Parameters
        ----------
        regions : MACS3.Signal.Region.Regions
            Sorted region collection whose intervals should be excluded.
        
        Notes
        -----
        The operation mutates the track in place and finishes by calling
        :meth:`finalize` to refresh cached statistics.
        """
        i: cython.ulong
        k: bytes
        locs: cnp.ndarray
        locs_size: cython.ulong
        chrnames: set
        regions_c: list
        selected_idx: cnp.ndarray
        regions_chrs: list
        r1: cnp.void            # this is the location in numpy.void -- like a tuple
        r2: tuple               # this is the region
        n_rl1: cython.long
        n_rl2: cython.long

        if not self.is_sorted:
            self.sort()

        assert isinstance(regions, Regions)
        regions.sort()
        regions_chrs = list(regions.regions.keys())

        chrnames = self.get_chr_names()

        for k in chrnames:      # for each chromosome
            locs = self.locations[k]
            locs_size = self.size[k]
            # let's check if k is in regions_chr
            if k not in regions_chrs:
                # do nothing and continue
                self.total += locs_size
                continue

            # discard overlapping reads and make a new locations[k]
            # initialize boolean array as all TRUE, or all being kept
            selected_idx = np.ones(locs_size, dtype=bool)

            regions_c = regions.regions[k]

            i = 0
            n_rl1 = locs_size   # the number of locations left
            n_rl2 = len(regions_c)  # the number of regions left
            rl1_k = iter(locs).__next__
            rl2_k = iter(regions_c).__next__
            r1 = rl1_k()        # take the first value
            n_rl1 -= 1          # remaining rl1
            r2 = rl2_k()
            n_rl2 -= 1          # remaining rl2
            while (True):
                # we do this until there is no r1 or r2 left.
                if r2[0] < r1[1] and r1[0] < r2[1]:
                    # since we found an overlap, r1 will be skipped/excluded
                    # and move to the next r1
                    # get rid of this one
                    n_rl1 -= 1
                    self.length -= cython.cast(cython.ulonglong, r1[1] - r1[0])
                    selected_idx[i] = False

                    if n_rl1 > 0:
                        r1 = rl1_k()  # take the next location
                        i += 1
                        continue
                    else:
                        break
                if r1[1] < r2[1]:
                    # in this case, we need to move to the next r1,
                    n_rl1 -= 1
                    if n_rl1 > 0:
                        r1 = rl1_k()  # take the next location
                        i += 1
                    else:
                        # no more r1 left
                        break
                else:
                    # in this case, we need to move the next r2
                    if n_rl2:
                        r2 = rl2_k()  # take the next region
                        n_rl2 -= 1
                    else:
                        # no more r2 left
                        break

            self.locations[k] = locs[selected_idx]
            self.size[k] = self.locations[k].shape[0]
            # free memory?
            # I know I should shrink it to 0 size directly,
            # however, on Mac OSX, it seems directly assigning 0
            # doesn't do a thing.
            selected_idx.resize(self.buffer_size, refcheck=False)
            selected_idx.resize(0, refcheck=False)
        self.finalize()         # use length to set average_template_length and total
        return

    @cython.boundscheck(False)  # do not check that np indices are valid
    @cython.ccall
    def filter_dup(self, maxnum: cython.int = -1):
        """Limit the number of duplicate fragments at identical coordinates.
        
        Parameters
        ----------
        maxnum : int, optional
            Maximum number of fragments allowed per unique ``(start, end)`` pair.
            A negative value disables duplicate filtering.
        
        Notes
        -----
        Fragments exceeding ``maxnum`` for the same coordinates are dropped and the
        aggregate template length is adjusted accordingly.

        Examples
        --------
        .. code-block:: python

            from MACS3.Signal.PairedEndTrack import PETrackI

            pe = PETrackI()
            pe.add_loc(b"chr1", 10, 40)
            pe.add_loc(b"chr1", 10, 40)
            pe.finalize()
            pe.filter_dup(maxnum=1)
        """
        n: cython.int
        loc_start: cython.int
        loc_end: cython.int
        current_loc_start: cython.int
        current_loc_end: cython.int
        i: cython.ulong
        locs_size: cython.ulong
        k: bytes
        locs: cnp.ndarray
        chrnames: set
        selected_idx: cnp.ndarray

        if maxnum < 0:
            return              # condition to return if not filtering

        if not self.is_sorted:
            self.sort()

        self.total = 0
        # self.length = 0
        self.average_template_length = 0.0

        chrnames = self.get_chr_names()

        for k in chrnames:      # for each chromosome
            locs = self.locations[k]
            locs_size = self.size[k]
            if locs_size == 1:
                # do nothing and continue
                self.total += locs_size
                continue
            # discard duplicate reads and make a new locations[k]
            # initialize boolean array as all TRUE, or all being kept
            selected_idx = np.ones(locs_size, dtype=bool)
            # get the first loc
            (current_loc_start, current_loc_end) = locs[0]
            i = 1               # index of new_locs
            n = 1  # the number of tags in the current genomic location
            for i in range(1, locs_size):
                (loc_start, loc_end) = locs[i]
                if loc_start != current_loc_start or loc_end != current_loc_end:
                    # not the same, update currnet_loc_start/end/l, reset n
                    current_loc_start = loc_start
                    current_loc_end = loc_end
                    n = 1
                    continue
                else:
                    # both ends are the same, add 1 to duplicate number n
                    n += 1
                    if n > maxnum:
                        # change the flag to False
                        selected_idx[i] = False
                        # subtract current_loc_l from self.length
                        self.length -= cython.cast(cython.ulonglong, current_loc_end - current_loc_start)
            self.locations[k] = locs[selected_idx]
            self.size[k] = self.locations[k].shape[0]
            self.total += self.size[k]
            # free memory?
            # I know I should shrink it to 0 size directly,
            # however, on Mac OSX, it seems directly assigning 0
            # doesn't do a thing.
            selected_idx.resize(self.buffer_size, refcheck=False)
            selected_idx.resize(0, refcheck=False)
        self.average_template_length = self.length / self.total
        return

    @cython.ccall
    def sample_percent(self,
                       percent: cython.float,
                       seed: cython.int = -1):
        """Down-sample fragments in place by a fixed percentage.
        
        Parameters
        ----------
        percent : float
            Fraction of fragments to keep per chromosome between 0 and 1 (inclusive).
        seed : int, optional
            Deterministic seed for the RNG; a negative value uses NumPy's global state.
        
        Notes
        -----
        Sampling is performed independently for each chromosome by shuffling the
        fragments, resizing the arrays, and restoring coordinate order.

        Examples
        --------
        .. code-block:: python

            from MACS3.Signal.PairedEndTrack import PETrackI

            pe = PETrackI()
            pe.add_loc(b"chr1", 10, 40)
            pe.add_loc(b"chr1", 50, 90)
            pe.finalize()
            pe.sample_percent(0.5, seed=123)
        """
        # num: number of reads allowed on a certain chromosome
        num: cython.uint
        k: bytes
        chrnames: set

        self.total = 0
        self.length = 0
        self.average_template_length = 0.0

        chrnames = self.get_chr_names()

        if seed >= 0:
            info(f"#   A random seed {seed} has been used")
            rs = np.random.RandomState(np.random.MT19937(np.random.SeedSequence(seed)))
            rs_shuffle = rs.shuffle
        else:
            rs_shuffle = np.random.shuffle

        for k in sorted(chrnames):
            # for each chromosome.
            # This loop body is too big, I may need to split code later...

            num = cython.cast(cython.uint,
                              round(self.locations[k].shape[0] * percent, 5))
            rs_shuffle(self.locations[k])
            self.locations[k].resize(num, refcheck=False)
            self.locations[k].sort(order=['l', 'r'])  # sort by leftmost positions
            self.size[k] = self.locations[k].shape[0]
            self.length += (self.locations[k]['r'] - self.locations[k]['l']).sum()
            self.total += self.size[k]
        self.average_template_length = cython.cast(cython.float, self.length)/self.total
        return

    @cython.ccall
    def sample_percent_copy(self,
                            percent: cython.float,
                            seed: cython.int = -1):
        """Return a down-sampled copy of the track.
        
        Parameters
        ----------
        percent : float
            Fraction of fragments to retain per chromosome between 0 and 1.
        seed : int, optional
            Deterministic seed used when shuffling; a negative value disables seeding.
        
        Returns
        -------
        PETrackI
            New track containing the sampled fragments with metadata copied over.

        Examples
        --------
        .. code-block:: python

            from MACS3.Signal.PairedEndTrack import PETrackI

            pe = PETrackI()
            pe.add_loc(b"chr1", 10, 40)
            pe.add_loc(b"chr1", 50, 90)
            pe.finalize()
            pe_subset = pe.sample_percent_copy(0.5, seed=123)
        """
        # num: number of reads allowed on a certain chromosome
        num: cython.uint
        k: bytes
        chrnames: set
        ret_petrackI: PETrackI
        loc: cnp.ndarray

        ret_petrackI = PETrackI(anno=self.annotation,
                                buffer_size=self.buffer_size)
        chrnames = self.get_chr_names()

        if seed >= 0:
            info(f"# A random seed {seed} has been used in the sampling function")
            rs = np.random.default_rng(seed)
        else:
            rs = np.random.default_rng()

        rs_shuffle = rs.shuffle

        # chrnames need to be sorted otherwise we can't assure reproducibility
        for k in sorted(chrnames):
            # for each chromosome.
            # This loop body is too big, I may need to split code later...
            loc = np.copy(self.locations[k])
            num = cython.cast(cython.uint, round(loc.shape[0] * percent, 5))
            rs_shuffle(loc)
            loc.resize(num, refcheck=False)
            loc.sort(order=['l', 'r'])  # sort by leftmost positions
            ret_petrackI.locations[k] = loc
            ret_petrackI.size[k] = loc.shape[0]
            ret_petrackI.length += (loc['r'] - loc['l']).sum()
            ret_petrackI.total += ret_petrackI.size[k]
        ret_petrackI.average_template_length = cython.cast(cython.float, ret_petrackI.length)/ret_petrackI.total
        ret_petrackI.set_rlengths(self.get_rlengths())
        return ret_petrackI

    @cython.ccall
    def sample_num(self,
                   samplesize: cython.ulong,
                   seed: cython.int = -1):
        """Down-sample fragments in place to approximately ``samplesize``.
        
        Parameters
        ----------
        samplesize : int
            Target number of fragments across all chromosomes.
        seed : int, optional
            Deterministic seed forwarded to :meth:`sample_percent`.
        
        Notes
        -----
        The method converts ``samplesize`` into a sampling fraction using ``self.total``.
        Ensure :meth:`finalize` has been called so counts are up to date.
        """
        percent: cython.float

        percent = cython.cast(cython.float, samplesize)/self.total
        self.sample_percent(percent, seed)
        return

    @cython.ccall
    def sample_num_copy(self,
                        samplesize: cython.ulong,
                        seed: cython.int = -1):
        """Return a down-sampled copy with approximately ``samplesize`` fragments.
        
        Parameters
        ----------
        samplesize : int
            Target number of fragments across all chromosomes.
        seed : int, optional
            Deterministic seed forwarded to :meth:`sample_percent_copy`.
        
        Returns
        -------
        PETrackI
            New track containing the sampled fragments.

        Examples
        --------
        .. code-block:: python

            from MACS3.Signal.PairedEndTrack import PETrackI

            pe = PETrackI()
            pe.add_loc(b"chr1", 10, 40)
            pe.add_loc(b"chr1", 50, 90)
            pe.finalize()
            pe_subset = pe.sample_num_copy(samplesize=1, seed=123)
        """
        percent: cython.float

        percent = cython.cast(cython.float, samplesize)/self.total
        return self.sample_percent_copy(percent, seed)

    @cython.ccall
    def print_to_bed(self, fhd=None):
        """Write fragments to a three-column BEDPE-style stream.
        
        Parameters
        ----------
        fhd : io.IOBase, optional
            Writable file-like object. Defaults to ``sys.stdout``.
        
        Notes
        -----
        Each fragment is emitted as ``chrom	start	end`` using decoded chromosome
        names and the stored integer coordinates.

        Examples
        --------
        .. code-block:: python

            import io
            from MACS3.Signal.PairedEndTrack import PETrackI

            pe = PETrackI()
            pe.add_loc(b"chr1", 10, 40)
            pe.finalize()
            buf = io.StringIO()
            pe.print_to_bed(buf)
        """
        i: cython.int
        s: cython.int
        e: cython.int
        k: bytes
        chrnames: set

        if not fhd:
            fhd = sys.stdout
        assert isinstance(fhd, io.IOBase)

        chrnames = self.get_chr_names()

        for k in chrnames:
            # for each chromosome.
            # This loop body is too big, I may need to split code later...

            locs = self.locations[k]

            for i in range(locs.shape[0]):
                s, e = locs[i]
                fhd.write("%s\t%d\t%d\n" % (k.decode(), s, e))
        return

    @cython.ccall
    def pileup_a_chromosome(self,
                            chrom: bytes,
                            scale_factor: cython.float = 1.0,
                            baseline_value: cython.float = 0.0) -> list:
        """Compute a coverage pileup for a single chromosome.
        
        Parameters
        ----------
        chrom : bytes
            Chromosome name to pile up.
        scale_factor : float, optional
            Value used to scale the resulting coverage.
        baseline_value : float, optional
            Minimum value enforced on the coverage array.
        
        Returns
        -------
        list
            Two-element list ``[positions, values]`` with numpy arrays describing
            the pileup breakpoints and scaled coverage.

        Examples
        --------
        .. code-block:: python

            from MACS3.Signal.PairedEndTrack import PETrackI

            pe = PETrackI()
            pe.add_loc(b"chr1", 10, 40)
            pe.add_loc(b"chr1", 50, 90)
            pe.finalize()
            p, v = pe.pileup_a_chromosome(b"chr1")
        """
        tmp_pileup: list

        tmp_pileup = quick_pileup(np.sort(self.locations[chrom]['l']),
                                  np.sort(self.locations[chrom]['r']),
                                  scale_factor, baseline_value)
        return tmp_pileup

    @cython.ccall
    def pileup_a_chromosome_c(self,
                              chrom: bytes,
                              ds: list,
                              scale_factor_s: list,
                              baseline_value: cython.float = 0.0) -> list:
        """Project paired-end fragments into pseudo single-end pileups.
        
        Parameters
        ----------
        chrom : bytes
            Chromosome name to pile up.
        ds : list[int]
            Fragment lengths used to build the projections.
        scale_factor_s : list[float]
            Scale factors paired with each entry in ``ds``.
        baseline_value : float, optional
            Minimum value enforced on the coverage array.
        
        Returns
        -------
        list
            Two-element list ``[positions, values]`` representing the merged pileup
            with the maximum value taken across projections.
        """
        tmp_pileup: list
        prev_pileup: list
        scale_factor: cython.float
        d: cython.long
        five_shift: cython.long
        three_shift: cython.long
        rlength: cython.long = self.get_rlengths()[chrom]

        if not self.is_sorted:
            self.sort()

        assert len(ds) == len(scale_factor_s), "ds and scale_factor_s must have the same length!"

        prev_pileup = None

        for i in range(len(scale_factor_s)):
            d = ds[i]
            scale_factor = scale_factor_s[i]
            five_shift = d//2
            three_shift = d//2

            tmp_pileup = se_all_in_one_pileup(self.locations[chrom]['l'],
                                              self.locations[chrom]['r'],
                                              five_shift,
                                              three_shift,
                                              rlength,
                                              scale_factor,
                                              baseline_value)

            if prev_pileup:
                prev_pileup = over_two_pv_array(prev_pileup,
                                                tmp_pileup,
                                                func="max")
            else:
                prev_pileup = tmp_pileup

        return prev_pileup

    @cython.ccall
    def pileup_bdg(self,
                   scale_factor: cython.float = 1.0,
                   baseline_value: cython.float = 0.0):
        """Build a ``bedGraphTrackI`` with pileups for every chromosome.
        
        Parameters
        ----------
        scale_factor : float, optional
            Value used to scale the coverage for each chromosome.
        baseline_value : float, optional
            Minimum value enforced on the coverage arrays.
        
        Returns
        -------
        bedGraphTrackI
            BedGraph track populated with per-chromosome pileup data.

        Examples
        --------
        .. code-block:: python

            from MACS3.Signal.PairedEndTrack import PETrackI

            pe = PETrackI()
            pe.add_loc(b"chr1", 10, 40)
            pe.finalize()
            bdg = pe.pileup_bdg()
        """
        tmp_pileup: list
        chrom: bytes
        bdg: bedGraphTrackI

        bdg = bedGraphTrackI(baseline_value=baseline_value)

        for chrom in sorted(self.get_chr_names()):
            tmp_pileup = quick_pileup(np.sort(self.locations[chrom]['l']),
                                      np.sort(self.locations[chrom]['r']),
                                      scale_factor,
                                      baseline_value)

            # save to bedGraph
            bdg.add_chrom_data(chrom,
                               pyarray('i', tmp_pileup[0]),
                               pyarray('f', tmp_pileup[1]))
        return bdg

    @cython.ccall
    def pileup_bdg_hmmr(self,
                        mapping: list,
                        baseline_value: cython.float = 0.0) -> list:
        """Generate HMMRATAC-style pileups for every chromosome.
        
        Parameters
        ----------
        mapping : list
            Weight mapping produced by HMMRATAC EM training describing the short,
            mono-, di-, and tri-nucleosomal signals.
        baseline_value : float, optional
            Reserved parameter for API compatibility; not currently applied.
        
        Returns
        -------
        list
            List of dictionaries mirroring ``mapping`` where each dictionary maps
            chromosome names to pileup arrays returned by
            :func:`pileup_from_LR_hmmratac`.

        Examples
        --------
        .. code-block:: python

            from MACS3.Signal.PairedEndTrack import PETrackI

            pe = PETrackI()
            pe.add_loc(b"chr1", 10, 40)
            pe.finalize()
            mapping = [{"short": 1.0, "mono": 0.0, "di": 0.0, "tri": 0.0}]
            pileups = pe.pileup_bdg_hmmr(mapping)
        """
        ret_pileup: list
        chroms: set
        chrom: bytes
        i: cython.int

        ret_pileup = []
        for i in range(len(mapping)):
            ret_pileup.append({})
        chroms = self.get_chr_names()
        for i in range(len(mapping)):
            for chrom in sorted(chroms):
                ret_pileup[i][chrom] = pileup_from_LR_hmmratac(self.locations[chrom], mapping[i])
        return ret_pileup


@cython.cclass
class PETrackII:
    """Paired-end track for single-cell ATAC fragments with barcode metadata.
    
    Each chromosome stores a structured array of fragment coordinates and counts
    alongside an integer-encoded barcode array to support barcode-aware analyses.

    Examples
    --------
    .. code-block:: python

        from MACS3.Signal.PairedEndTrack import PETrackII

        pe = PETrackII(anno="scatac")
        pe.add_loc(b"chr1", 10, 40, barcode=b"BC01", count=1)
        pe.add_loc(b"chr1", 50, 90, barcode=b"BC02", count=2)
        pe.finalize()
        pe.set_rlengths({b"chr1": 1000})
    """
    locations = cython.declare(dict, visibility="public")
    # add another dict for storing barcode for each fragment we will
    # first convert barcode into integer and remember them in the
    # barcode_dict, which will map the rule to numerize
    # key:bytes as value:4bytes_integer
    barcodes = cython.declare(dict, visibility="public")
    barcode_dict = cython.declare(dict, visibility="public")
    # the last number for barcodes, used to map barcode to integer
    barcode_last_n: cython.int

    size = cython.declare(dict, visibility="public")
    buf_size = cython.declare(dict, visibility="public")
    is_sorted = cython.declare(bool, visibility="public")
    total = cython.declare(cython.ulong, visibility="public")
    annotation = cython.declare(str, visibility="public")
    # rlengths: reference chromosome lengths dictionary
    rlengths = cython.declare(dict, visibility="public")
    buffer_size = cython.declare(cython.long, visibility="public")
    length = cython.declare(cython.ulonglong, visibility="public")  # total length of all fragments
    average_template_length = cython.declare(cython.float, visibility="public")
    is_destroyed: bool

    def __init__(self, anno: str = "", buffer_size: cython.long = 100000):
        # dictionary with chrname as key, nparray with
        # [('l','i4'),('r','i4'),('c','u2')] as value
        self.locations = {}
        # dictionary with chrname as key, size of the above nparray as value
        # size is to remember the size of the fragments added to this chromosome
        self.size = {}
        # dictionary with chrname as key, size of the above nparray as value
        self.buf_size = {}
        self.is_sorted = False
        self.total = 0           # total fragments
        self.annotation = anno   # need to be figured out
        self.rlengths = {}
        self.buffer_size = buffer_size
        self.length = 0
        self.average_template_length = 0.0
        self.is_destroyed = False

        self.barcodes = {}
        self.barcode_dict = {}
        self.barcode_last_n = 0

    @cython.ccall
    def add_loc(self,
                chromosome: bytes,
                start: cython.int,
                end: cython.int,
                barcode: bytes,
                count: cython.ushort):
        """Append a fragment together with its barcode and count.
        
        Parameters
        ----------
        chromosome : bytes
            Chromosome name (as bytes) for the fragment.
        start : int
            Zero-based start coordinate of the fragment.
        end : int
            Zero-based end coordinate of the fragment.
        barcode : bytes
            Raw barcode sequence associated with the fragment.
        count : int
            Number of occurrences represented by the fragment.
        
        Notes
        -----
        Barcodes are interned into integers via ``barcode_dict`` for compact storage
        and the accumulated template length is weighted by ``count``.

        Examples
        --------
        .. code-block:: python

            from MACS3.Signal.PairedEndTrack import PETrackII

            pe = PETrackII()
            pe.add_loc(b"chr1", 10, 40, barcode=b"BC01", count=1)
            pe.add_loc(b"chr1", 50, 90, barcode=b"BC02", count=2)
        """
        i: cython.int
        # bn: the integer in barcode_dict for this barcode
        bn: cython.int

        if barcode not in self.barcode_dict:
            self.barcode_dict[barcode] = self.barcode_last_n
            self.barcode_last_n += 1
        bn = self.barcode_dict[barcode]

        if chromosome not in self.locations:
            self.buf_size[chromosome] = self.buffer_size
            # note: ['l'] is the leftmost end, ['r'] is the rightmost end of fragment.
            # ['c'] is the count number of this fragment
            self.locations[chromosome] = np.zeros(shape=self.buffer_size,
                                                  dtype=[('l', 'i4'), ('r', 'i4'), ('c', 'u2')])
            self.barcodes[chromosome] = np.zeros(shape=self.buffer_size,
                                                 dtype='i4')
            self.locations[chromosome][0] = (start, end, count)
            self.barcodes[chromosome][0] = bn
            self.size[chromosome] = 1
        else:
            i = self.size[chromosome]
            if self.buf_size[chromosome] == i:
                self.buf_size[chromosome] += self.buffer_size
                self.locations[chromosome].resize((self.buf_size[chromosome]),
                                                  refcheck=False)
                self.barcodes[chromosome].resize((self.buf_size[chromosome]),
                                                 refcheck=False)                
            self.locations[chromosome][i] = (start, end, count)
            self.barcodes[chromosome][i] = bn
            self.size[chromosome] = i + 1
        self.length += (end - start) * count
        return

    @cython.ccall
    def destroy(self):
        """Release fragment and barcode arrays held by the track.
        
        All per-chromosome arrays are resized to zero, barcode mappings are cleared,
        and the track is marked as destroyed.
        """
        chrs: set
        chromosome: bytes

        chrs = self.get_chr_names()
        for chromosome in sorted(chrs):
            if chromosome in self.locations:
                self.locations[chromosome].resize(self.buffer_size,
                                                  refcheck=False)
                self.locations[chromosome].resize(0,
                                                  refcheck=False)
                self.locations[chromosome] = None
                self.locations.pop(chromosome)
                self.barcodes.resize(self.buffer_size,
                                     refcheck=False)
                self.barcodes.resize(0,
                                     refcheck=False)
                self.barcodes[chromosome] = None
                self.barcodes.pop(chromosome)
        self.barcode_dict = {}
        self.is_destroyed = True
        return

    @cython.ccall
    def set_rlengths(self, rlengths: dict) -> bool:
        """Attach reference chromosome lengths to the track.
        
        Parameters
        ----------
        rlengths : dict
            Mapping from chromosome name (bytes) to reference length.
        
        Returns
        -------
        bool
            True when the length mapping has been updated.
        
        Notes
        -----
        Any chromosome stored in the track but missing from ``rlengths`` is assigned
        ``INT_MAX`` so downstream bounds checks can succeed.
        """
        valid_chroms: set
        missed_chroms: set
        chrom: bytes

        valid_chroms = set(self.locations.keys()).intersection(rlengths.keys())
        for chrom in sorted(valid_chroms):
            self.rlengths[chrom] = rlengths[chrom]
        missed_chroms = set(self.locations.keys()).difference(rlengths.keys())
        for chrom in sorted(missed_chroms):
            self.rlengths[chrom] = INT_MAX
        return True

    @cython.ccall
    def get_rlengths(self) -> dict:
        """Return the reference chromosome lengths associated with the track.
        
        Returns
        -------
        dict
            Mapping from chromosome name (bytes) to reference length. Chromosomes
            without a recorded length default to ``INT_MAX``.
        """
        if not self.rlengths:
            self.rlengths = dict([(k, INT_MAX) for k in self.locations.keys()])
        return self.rlengths

    @cython.ccall
    def finalize(self):
        """Shrink arrays, sort fragments, and refresh aggregate counters.
        
        Each per-chromosome fragment array is resized to its observed length, sorted
        by ``('l', 'r')``, and the accompanying barcode array is reordered to match.
        The method updates ``total`` and ``average_template_length`` using count
        weights and marks the track as sorted.
        
        Raises
        ------
        AssertionError
            If no fragments are present when finalizing.

        Examples
        --------
        .. code-block:: python

            from MACS3.Signal.PairedEndTrack import PETrackII

            pe = PETrackII()
            pe.add_loc(b"chr1", 10, 40, barcode=b"BC01", count=1)
            pe.finalize()
        """
        c: bytes
        chrnames: set
        indices: cnp.ndarray

        self.total = 0

        chrnames = self.get_chr_names()

        for c in chrnames:
            self.locations[c].resize((self.size[c]), refcheck=False)
            if self.size[c] == 0:
                if c in self.size:
                    del self.size[c]
                if c in self.locations:
                    del self.locations[c]
                if c in self.barcodes:
                    del self.barcodes[c]
                if c in self.rlengths:
                    del self.rlengths[c]
                continue
            indices = np.argsort(self.locations[c], order=['l', 'r'])
            self.locations[c] = self.locations[c][indices]
            self.barcodes[c] = self.barcodes[c][indices]
            self.total += np.sum(self.locations[c]['c'])  # self.size[c]

        assert self.total > 0, "Error: no fragments in PETrackII"

        self.is_sorted = True
        self.average_template_length = cython.cast(cython.float,
                                                   self.length) / self.total
        return

    @cython.ccall
    def get_locations_by_chr(self, chromosome: bytes):
        """Return the fragment array for a chromosome.
        
        Parameters
        ----------
        chromosome : bytes
            Chromosome name, provided as bytes.
        
        Returns
        -------
        numpy.ndarray
            Structured array with ``('l', 'i4')``, ``('r', 'i4')``, and ``('c', 'u2')`` fields.
        
        Raises
        ------
        Exception
            If the chromosome is not present in the track.
        """
        if chromosome in self.locations:
            return self.locations[chromosome]
        else:
            raise Exception("No such chromosome name (%s) in TrackI object!\n" % (chromosome))

    @cython.ccall
    def get_chr_names(self) -> set:
        """Return the set of chromosome names stored in the track.
        
        Returns
        -------
        set
            Chromosome names (bytes) that currently have fragments.
        """
        return set(self.locations.keys())

    @cython.ccall
    def sort(self):
        """Sort fragments and barcodes for each chromosome.
        
        Fragments are ordered first by their left coordinate and then by their right
        coordinate, and the barcode array is reordered alongside the fragment array.
        The ``is_sorted`` flag is set to ``True`` when sorting completes.
        """
        c: bytes
        chrnames: set
        indices: cnp.ndarray

        chrnames = self.get_chr_names()

        for c in chrnames:
            indices = np.argsort(self.locations[c], order=['l', 'r'])
            self.locations[c] = self.locations[c][indices]
            self.barcodes[c] = self.barcodes[c][indices]
        self.is_sorted = True
        return

    @cython.ccall
    def count_fraglengths(self) -> dict:
        """Count fragment lengths weighted by per-fragment counts.
        
        Returns
        -------
        dict
            Mapping from fragment length to the total count contributed by fragments
            of that length.
        """
        sizes: cnp.ndarray(cnp.int32_t, ndim=1)
        s: cython.int
        locs: cnp.ndarray
        chrnames: list
        i: cython.int
        j: cython.int

        counter = Counter()
        chrnames = list(self.get_chr_names())
        for i in range(len(chrnames)):
            locs = self.locations[chrnames[i]]
            sizes = locs['r'] - locs['l']
            for j in range(len(sizes)):
                s = sizes[j]
                counter[s] += locs['c'][j]
        return dict(counter)

    @cython.ccall
    def fraglengths(self) -> cnp.ndarray:
        """Return all fragment lengths expanded by their counts.
        
        Returns
        -------
        numpy.ndarray
            Array of ``end - start`` values repeated according to the stored counts.
        """
        sizes: cnp.ndarray(np.int32_t, ndim=1)
        chrnames: list
        out: list
        chrom: bytes

        chrnames = list(self.get_chr_names())
        out = []
        for chrom in chrnames:
            locs = self.locations[chrom]
            sizes = locs['r'] - locs['l']
            counts = locs['c']
            out.append(np.repeat(sizes, counts))
        if out:
            return np.concatenate(out).astype(np.int32)
        else:
            return np.array([], dtype=np.int32)

    @cython.ccall
    def subset(self, selected_barcodes: set):
        """Build a new track containing only fragments from selected barcodes.
        
        Parameters
        ----------
        selected_barcodes : set
            Set of barcode byte strings to retain.
        
        Returns
        -------
        PETrackII
            New track restricted to the provided barcodes with metadata preserved.

        Examples
        --------
        .. code-block:: python

            from MACS3.Signal.PairedEndTrack import PETrackII

            pe = PETrackII()
            pe.add_loc(b"chr1", 10, 40, barcode=b"BC01", count=1)
            pe.add_loc(b"chr1", 50, 90, barcode=b"BC02", count=1)
            pe.finalize()
            subset = pe.subset({b"BC01"})
        """
        indices: cnp.ndarray
        chrs: set
        selected_barcodes_filtered: list
        selected_barcodes_n: list
        chromosome: bytes
        ret: PETrackII

        ret = PETrackII()
        chrs = self.get_chr_names()

        # first we need to convert barcodes into integers in our
        # barcode_dict
        selected_barcodes_filtered = [b
                                      for b in selected_barcodes
                                      if b in self.barcode_dict]
        ret.barcode_dict = {b: self.barcode_dict[b]
                            for b in selected_barcodes_filtered}
        selected_barcodes_n = [self.barcode_dict[b]
                               for b in selected_barcodes_filtered]
        ret.barcode_last_n = self.barcode_last_n

        # pass some values from self to ret
        ret.annotation = self.annotation
        ret.is_sorted = self.is_sorted
        ret.rlengths = self.rlengths
        ret.buffer_size = self.buffer_size
        # ret.total = 0
        ret.length = 0
        # ret.average_template_length = 0
        ret.is_destroyed = True

        for chromosome in sorted(chrs):
            indices = np.where(np.isin(self.barcodes[chromosome],
                                       list(selected_barcodes_n)))[0]
            ret.barcodes[chromosome] = self.barcodes[chromosome][indices]
            ret.locations[chromosome] = self.locations[chromosome][indices]
            ret.size[chromosome] = len(ret.locations[chromosome])
            ret.buf_size[chromosome] = ret.size[chromosome]
            # ret.total += np.sum(ret.locations[chromosome]['c'])
            ret.length += np.sum((ret.locations[chromosome]['r'] -
                                  ret.locations[chromosome]['l']) *
                                 ret.locations[chromosome]['c'])
        ret.finalize()
        # ret.average_template_length = ret.length / ret.total
        return ret

    @cython.ccall
    def pileup_a_chromosome(self,
                            chrom: bytes,
                            scale_factor: cython.float = 1.0,
                            baseline_value: cython.float = 0.0) -> list:
        """Compute a coverage pileup for a single chromosome.
        
        Parameters
        ----------
        chrom : bytes
            Chromosome name to pile up.
        scale_factor : float, optional
            Value used to scale the resulting coverage.
        baseline_value : float, optional
            Minimum value enforced on the coverage array.
        
        Returns
        -------
        list
            Two-element list ``[positions, values]`` with numpy arrays describing
            the pileup breakpoints and scaled coverage.
        """
        pv: cnp.ndarray
        v: cnp.ndarray

        pv = pileup_from_LRC(self.locations[chrom])
        v = pv['v']
        v = v * scale_factor
        v[v < baseline_value] = baseline_value

        return [pv['p'], v]

    @cython.ccall
    def pileup_a_chromosome_c(self,
                              chrom: bytes,
                              ds: list,
                              scale_factor_s: list,
                              baseline_value: cython.float = 0.0) -> list:
        """Project paired-end fragments into pseudo single-end pileups.
        
        Parameters
        ----------
        chrom : bytes
            Chromosome name to pile up.
        ds : list[int]
            Fragment lengths used to build the projections.
        scale_factor_s : list[float]
            Scale factors paired with each entry in ``ds``.
        baseline_value : float, optional
            Minimum value enforced on the coverage array.
        
        Returns
        -------
        list
            Two-element list ``[positions, values]`` representing the merged pileup
            with the maximum value taken across projections.
        """
        prev_pileup: list
        scale_factor: cython.float
        d: cython.long
        five_shift: cython.long

        pv: cnp.ndarray
        v: cnp.ndarray
        tmp_arr_l: cnp.ndarray
        tmp_arr_r: cnp.ndarray
        tmp_arr: cnp.ndarray

        ####
        if not self.is_sorted:
            self.sort()

        assert len(ds) == len(scale_factor_s), "ds and scale_factor_s must have the same length!"

        prev_pileup = None

        for i in range(len(scale_factor_s)):
            d = ds[i]
            scale_factor = scale_factor_s[i]
            five_shift = d//2

            # note, we have to pileup left ends and right ends separately
            tmp_arr_l = self.locations[chrom].copy()
            tmp_arr_l['l'] = tmp_arr_l['l'] - five_shift
            tmp_arr_l['r'] = tmp_arr_l['l'] + d

            tmp_arr_r = self.locations[chrom].copy()
            tmp_arr_r['l'] = tmp_arr_r['r'] - five_shift
            tmp_arr_r['r'] = tmp_arr_r['l'] + d

            tmp_arr = np.concatenate([tmp_arr_l, tmp_arr_r])
            del tmp_arr_l
            del tmp_arr_r

            pv = pileup_from_LRC(tmp_arr)

            v = pv['v']
            v = v * scale_factor
            v[v < baseline_value] = baseline_value

            if prev_pileup:
                prev_pileup = over_two_pv_array(prev_pileup,
                                                [pv['p'], v],
                                                func="max")
            else:
                prev_pileup = [pv['p'], v]

        return prev_pileup

    @cython.ccall
    def pileup_bdg(self,
                   scale_factor: cython.float = 1.0,
                   baseline_value: cython.float = 0.0):
        """Build a ``bedGraphTrackI`` with pileups for every chromosome.
        
        Parameters
        ----------
        scale_factor : float, optional
            Value used to scale the coverage for each chromosome.
        baseline_value : float, optional
            Minimum value enforced on the coverage arrays.
        
        Returns
        -------
        bedGraphTrackI
            BedGraph track populated with per-chromosome pileup data.
        """
        bdg: bedGraphTrackI
        pv: cnp.ndarray
        chrom: bytes

        bdg = bedGraphTrackI(baseline_value=baseline_value)
        for chrom in sorted(self.get_chr_names()):
            pv = pileup_from_LRC(self.locations[chrom])
            v = pv['v']
            v = v * scale_factor
            v[v < baseline_value] = baseline_value
            bdg.add_chrom_data(chrom,
                               pyarray('i', pv['p']),
                               pyarray('f', v))
        return bdg

    @cython.ccall
    def pileup_bdg2(self):
        """Build a ``bedGraphTrackII`` with pileups for every chromosome.
        
        Returns
        -------
        bedGraphTrackII
            BedGraph track populated with per-chromosome pileup arrays and finalized.
        """
        bdg: bedGraphTrackII
        pv: cnp.ndarray

        bdg = bedGraphTrackII()
        for chrom in self.get_chr_names():
            pv = pileup_from_LRC(self.locations[chrom])
            bdg.add_chrom_data(chrom, pv)
        # bedGraphTrackII needs to be 'finalized'.
        bdg.finalize()
        return bdg

    @cython.boundscheck(False)  # do not check that np indices are valid
    @cython.ccall
    def exclude(self, regions):
        """Remove fragments that overlap the provided exclusion regions.
        
        Parameters
        ----------
        regions : MACS3.Signal.Region.Regions
            Sorted region collection whose intervals should be excluded.
        
        Notes
        -----
        The operation mutates the track in place, adjusts fragment counts and lengths,
        and finishes by calling :meth:`finalize`.
        """
        i: cython.ulong
        k: bytes
        locs: cnp.ndarray
        locs_size: cython.ulong
        chrnames: set
        regions_c: list
        selected_idx: cnp.ndarray
        regions_chrs: list
        r1: cnp.void
        r2: tuple
        n_rl1: cython.long
        n_rl2: cython.long

        if not self.is_sorted:
            self.sort()

        assert isinstance(regions, Regions)
        regions.sort()
        regions_chrs = list(regions.regions.keys())

        chrnames = self.get_chr_names()

        for k in chrnames:      # for each chromosome
            locs = self.locations[k]
            locs_size = self.size[k]
            # let's check if k is in regions_chr
            if k not in regions_chrs:
                # do nothing and continue
                continue

            # discard overlapping reads and make a new locations[k]
            # initialize boolean array as all TRUE, or all being kept
            selected_idx = np.ones(locs_size, dtype=bool)

            regions_c = regions.regions[k]

            i = 0
            n_rl1 = len(locs)
            n_rl2 = len(regions_c)
            rl1_k = iter(locs).__next__
            rl2_k = iter(regions_c).__next__
            r1 = rl1_k()
            n_rl1 -= 1          # remaining rl1
            r2 = rl2_k()
            n_rl2 -= 1          # remaining rl2
            while (True):
                # we do this until there is no r1 or r2 left.
                if r2[0] < r1[1] and r1[0] < r2[1]:
                    # since we found an overlap, r1 will be skipped/excluded
                    # and move to the next r1
                    # get rid of this one
                    n_rl1 -= 1
                    self.length -= cython.cast(cython.ulonglong, (r1[1] - r1[0])*r1[2])
                    selected_idx[i] = False

                    if n_rl1 >= 0:
                        r1 = rl1_k()
                        i += 1
                        continue
                    else:
                        break
                if r1[1] < r2[1]:
                    # in this case, we need to move to the next r1,
                    n_rl1 -= 1
                    if n_rl1 >= 0:
                        r1 = rl1_k()
                        i += 1
                    else:
                        # no more r1 left
                        break
                else:
                    # in this case, we need to move the next r2
                    if n_rl2:
                        r2 = rl2_k()
                        n_rl2 -= 1
                    else:
                        # no more r2 left
                        break

            self.locations[k] = locs[selected_idx]
            self.barcodes[k] = self.barcodes[k][selected_idx]
            self.size[k] = self.locations[k].shape[0]
            # free memory?
            # I know I should shrink it to 0 size directly,
            # however, on Mac OSX, it seems directly assigning 0
            # doesn't do a thing.
            selected_idx.resize(self.buffer_size, refcheck=False)
            selected_idx.resize(0, refcheck=False)
        self.finalize()
        return

    @cython.ccall
    def sample_percent(self,
                       percent: cython.float,
                       seed: cython.int = -1):
        """Down-sample fragments in place so counts reflect a given percentage.
        
        Parameters
        ----------
        percent : float
            Fraction of total counts to keep per chromosome between 0 and 1 (inclusive).
        seed : int, optional
            Deterministic seed for the RNG; a negative value uses NumPy's global state.
        
        Notes
        -----
        Fragments are sampled proportionally to their counts by expanding to an index
        vector, shuffling, and collapsing counts for the retained entries. Aggregate
        statistics are recomputed and the result is resorted.
        """
        k: bytes
        loc: cnp.ndarray
        bar: cnp.ndarray
        counts: cnp.ndarray
        n: cython.uint
        n_sample: cython.uint
        idx_flat: cnp.ndarray
        unique_idx: cnp.ndarray
        new_counts: cnp.ndarray
        new_locs: cnp.ndarray
        new_bars: cnp.ndarray

        assert 0.0 <= percent <= 1.0, "percent must be in [0, 1]"
        chrnames = sorted(self.get_chr_names())

        # Setup shuffling logic like PETrackI
        if seed >= 0:
            info(f"#   A random seed {seed} has been used")
            rs = np.random.RandomState(np.random.MT19937(np.random.SeedSequence(seed)))
            rs_shuffle = rs.shuffle
        else:
            rs_shuffle = np.random.shuffle

        self.length = 0
        self.total = 0
        self.average_template_length = 0.0

        for k in chrnames:
            loc = self.locations[k]
            bar = self.barcodes[k]
            counts = loc['c']
            n = int(counts.sum())
            n_sample = int(round(n * percent))
            if n == 0 or n_sample == 0:
                self.locations[k] = loc[:0]
                self.barcodes[k] = bar[:0]
                self.size[k] = 0
                continue

            # Flatten: build an array of indices into loc, repeated by count
            idx_flat = np.repeat(np.arange(len(loc)), counts)
            rs_shuffle(idx_flat)
            idx_flat = idx_flat[:n_sample]

            # Recount: count how many times each index is chosen
            unique_idx, new_counts = np.unique(idx_flat, return_counts=True)
            # Compose new arrays
            new_locs = loc[unique_idx].copy()
            new_locs['c'] = new_counts
            new_bars = bar[unique_idx].copy()
            self.locations[k] = new_locs
            self.barcodes[k] = new_bars
            self.size[k] = len(new_locs)
            self.length += np.sum((new_locs['r'] - new_locs['l']) * new_locs['c'])
            self.total += np.sum(new_locs['c'])

        if self.total > 0:
            self.average_template_length = float(self.length) / self.total
        else:
            self.average_template_length = 0.0
        self.sort()
        return

    @cython.ccall
    def sample_percent_copy(self,
                            percent: cython.float,
                            seed: cython.int = -1):
        """Return a down-sampled copy whose counts reflect a given percentage.
        
        Parameters
        ----------
        percent : float
            Fraction of total counts to keep per chromosome between 0 and 1 (inclusive).
        seed : int, optional
            Deterministic seed for the RNG; a negative value uses NumPy's global state.
        
        Returns
        -------
        PETrackII
            New track containing the sampled fragments with metadata preserved.
        
        Notes
        -----
        Fragments are sampled proportionally to their counts and the returned track is
        sorted with reference lengths copied from the source track.
        """
        k: bytes
        loc: cnp.ndarray
        bar: cnp.ndarray
        counts: cnp.ndarray
        n: cython.uint
        n_sample: cython.uint
        idx_flat: cnp.ndarray
        unique_idx: cnp.ndarray
        new_counts: cnp.ndarray
        new_locs: cnp.ndarray
        new_bars: cnp.ndarray

        assert 0.0 <= percent <= 1.0, "percent must be in [0, 1]"
        chrnames = sorted(self.get_chr_names())

        # Setup shuffling logic like PETrackI
        if seed >= 0:
            info(f"#   A random seed {seed} has been used")
            rs = np.random.RandomState(np.random.MT19937(np.random.SeedSequence(seed)))
            rs_shuffle = rs.shuffle
        else:
            rs_shuffle = np.random.shuffle

        ret = PETrackII(anno=self.annotation, buffer_size=self.buffer_size)
        ret.barcode_dict = dict(self.barcode_dict)
        ret.barcode_last_n = self.barcode_last_n

        ret.length = 0
        ret.total = 0

        for k in chrnames:
            loc = self.locations[k]
            bar = self.barcodes[k]
            counts = loc['c']
            n = int(counts.sum())
            n_sample = int(round(n * percent))
            if n == 0 or n_sample == 0:
                ret.locations[k] = loc[:0]
                ret.barcodes[k] = bar[:0]
                ret.size[k] = 0
                ret.buf_size[k] = 0
                continue

            idx_flat = np.repeat(np.arange(len(loc)), counts)
            rs_shuffle(idx_flat)
            idx_flat = idx_flat[:n_sample]
            unique_idx, new_counts = np.unique(idx_flat, return_counts=True)
            new_locs = loc[unique_idx].copy()
            new_locs['c'] = new_counts
            new_bars = bar[unique_idx].copy()
            ret.locations[k] = new_locs
            ret.barcodes[k] = new_bars
            ret.size[k] = len(new_locs)
            ret.buf_size[k] = len(new_locs)
            ret.length += np.sum((new_locs['r'] - new_locs['l']) * new_locs['c'])
            ret.total += np.sum(new_locs['c'])

        if ret.total > 0:
            ret.average_template_length = float(ret.length) / ret.total
        else:
            ret.average_template_length = 0.0
        ret.set_rlengths(self.get_rlengths())
        ret.sort()
        return ret

    @cython.ccall
    def sample_num(self,
                   samplesize: cython.ulong,
                   seed: cython.int = -1):
        """Down-sample fragments in place so total counts approximate ``samplesize``.
        
        Parameters
        ----------
        samplesize : int
            Target total count across all chromosomes.
        seed : int, optional
            Deterministic seed forwarded to :meth:`sample_percent`.
        
        Notes
        -----
        The method converts ``samplesize`` into a sampling fraction using the current
        total count and reuses :meth:`sample_percent`.
        """
        chrnames: set
        n_total: cython.uint
        chr_totals: dict
        k: bytes
        percent: cython.float

        chrnames = self.get_chr_names()
        n_total = 0
        chr_totals = {}
        for k in chrnames:
            chr_totals[k] = self.locations[k]['c'].sum()
            n_total += chr_totals[k]
        percent = 0.0 if n_total == 0 else min(samplesize / n_total, 1.0)
        self.sample_percent(percent, seed)
        return

    @cython.ccall
    def sample_num_copy(self,
                        samplesize: cython.ulong,
                        seed: cython.int = -1):
        """Return a down-sampled copy whose total counts approximate ``samplesize``.
        
        Parameters
        ----------
        samplesize : int
            Target total count across all chromosomes.
        seed : int, optional
            Deterministic seed forwarded to :meth:`sample_percent_copy`.
        
        Returns
        -------
        PETrackII
            New track containing the sampled fragments.
        """
        chrnames: set
        n_total: cython.uint
        chr_totals: dict
        k: bytes
        percent: cython.float

        chrnames = self.get_chr_names()
        n_total = 0
        chr_totals = {}
        for k in chrnames:
            chr_totals[k] = self.locations[k]['c'].sum()
            n_total += chr_totals[k]
        percent = 0.0 if n_total == 0 else min(samplesize / n_total, 1.0)
        return self.sample_percent_copy(percent, seed)

    @cython.ccall
    def pileup_bdg_hmmr(self,
                        mapping: list,
                        baseline_value: cython.float = 0.0) -> list:
        """Generate HMMRATAC-style pileups for every chromosome.
        
        Parameters
        ----------
        mapping : list
            Weight mapping produced by HMMRATAC EM training describing the short,
            mono-, di-, and tri-nucleosomal signals.
        baseline_value : float, optional
            Reserved parameter for API compatibility; not currently applied.
        
        Returns
        -------
        list
            List of dictionaries mirroring ``mapping`` where each dictionary maps
            chromosome names to pileup arrays returned by
            :func:`pileup_from_LR_hmmratac`.
        """
        ret_pileup: list
        i: cython.uint
        chroms: set
        chrom: bytes
        locs: cnp.ndarray
        counts: cnp.ndarray
        LR_expanded: cnp.ndarray
        idx: cnp.ndarray

        ret_pileup = []
        for i in range(len(mapping)):
            ret_pileup.append({})

        chroms = self.get_chr_names()
        for i in range(len(mapping)):
            for chrom in sorted(chroms):
                locs = self.locations[chrom]
                counts = locs['c']
                # Efficient numpy "explode"
                if locs.shape[0] == 0 or counts.sum() == 0:
                    LR_expanded = np.zeros((0,),
                                           dtype=[('l', 'i4'), ('r', 'i4')])
                else:
                    idx = np.repeat(np.arange(locs.shape[0]), counts)
                    LR_expanded = np.empty((len(idx),),
                                           dtype=[('l', 'i4'), ('r', 'i4')])
                    LR_expanded['l'] = locs['l'][idx]
                    LR_expanded['r'] = locs['r'][idx]
                ret_pileup[i][chrom] = pileup_from_LR_hmmratac(LR_expanded, mapping[i])
        return ret_pileup



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2025-11-10 15:24:27 Tao Liu>

"""Module for FWTrack classes.

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

# ------------------------------------
# python modules
# ------------------------------------
import sys
import io

# ------------------------------------
# MACS3 modules
# ------------------------------------

from MACS3.IO.PeakIO import PeakIO
from MACS3.Signal.Pileup import se_all_in_one_pileup, over_two_pv_array

# ------------------------------------
# Other modules
# ------------------------------------
import cython
import numpy as np
from cython.cimports.cpython import bool
import cython.cimports.numpy as cnp
from cython.cimports.libc.stdint import INT32_MAX as INT_MAX

# ------------------------------------
# constants
# ------------------------------------

# ------------------------------------
# Misc functions
# ------------------------------------

# ------------------------------------
# Classes
# ------------------------------------


@cython.cclass
class FWTrack:
    """Fixed-width fragment track grouped by chromosome.
    
    Stores plus- and minus-strand 5' cut positions in numpy arrays and exposes
    utilities for sorting, filtering, sampling, and pileup generation.
    """
    locations: dict
    pointer: dict
    buf_size: dict
    rlengths: dict
    is_sorted: bool
    is_destroyed: bool
    total = cython.declare(cython.ulong, visibility="public")
    annotation = cython.declare(str, visibility="public")
    buffer_size = cython.declare(cython.long, visibility="public")
    length = cython.declare(cython.ulonglong, visibility="public")
    fw = cython.declare(cython.int, visibility="public")

    def __init__(self,
                 fw: cython.int = 0,
                 anno: str = "",
                 buffer_size: cython.long = 100000):
        """Initialize an empty fixed-width track.
        
        Parameters
        ----------
        fw : int, optional
            Fixed fragment width (bp) used when estimating coverage and region length.
        anno : str, optional
            Annotation label retained with the track metadata.
        buffer_size : int, optional
            Number of positions allocated per growth chunk for each strand array.
        """
        self.fw = fw
        self.locations = {}    # location pairs: two strands
        self.pointer = {}      # location pairs
        self.buf_size = {}     # location pairs
        self.is_sorted = False
        self.total = 0           # total tags
        self.annotation = anno   # need to be figured out
        # lengths of reference sequences, e.g. each chromosome in a genome
        self.rlengths = {}
        self.buffer_size = buffer_size
        self.length = 0
        self.is_destroyed = False

    @cython.ccall
    def destroy(self):
        """Release numpy buffers held by the track.
        
        All per-chromosome arrays are resized to zero so the memory footprint returns
        to the allocator, and the track is marked as destroyed.
        """
        chrs: set
        chromosome: bytes

        chrs = self.get_chr_names()
        for chromosome in sorted(chrs):
            if chromosome in self.locations:
                self.locations[chromosome][0].resize(self.buffer_size,
                                                     refcheck=False)
                self.locations[chromosome][0].resize(0,
                                                     refcheck=False)
                self.locations[chromosome][1].resize(self.buffer_size,
                                                     refcheck=False)
                self.locations[chromosome][1].resize(0,
                                                     refcheck=False)
                self.locations[chromosome] = [None, None]
                self.locations.pop(chromosome)
        self.is_destroyed = True
        return

    @cython.ccall
    def add_loc(self,
                chromosome: bytes,
                fiveendpos: cython.int,
                strand: cython.int):
        """Append a 5' cut position to the track.
        
        Parameters
        ----------
        chromosome : bytes
            Chromosome name (as bytes) that owns the cut.
        fiveendpos : int
            Zero-based 5' coordinate of the cut site.
        strand : int
            Strand flag where ``0`` denotes plus and ``1`` denotes minus.
        
        Notes
        -----
        Positions are stored in strand-specific numpy arrays keyed by chromosome, and
        the strand pointer is advanced as new positions are appended.
        """
        i: cython.int
        b: cython.int
        arr: cnp.ndarray

        if chromosome not in self.locations:
            self.buf_size[chromosome] = [self.buffer_size, self.buffer_size]
            self.locations[chromosome] = [np.zeros(self.buffer_size, dtype='i4'),
                                          np.zeros(self.buffer_size, dtype='i4')]
            self.pointer[chromosome] = [0, 0]
            self.locations[chromosome][strand][0] = fiveendpos
            self.pointer[chromosome][strand] = 1
        else:
            i = self.pointer[chromosome][strand]
            b = self.buf_size[chromosome][strand]
            arr = self.locations[chromosome][strand]
            if b == i:
                b += self.buffer_size
                arr.resize(b, refcheck=False)
                self.buf_size[chromosome][strand] = b
            arr[i] = fiveendpos
            self.pointer[chromosome][strand] += 1
        return

    @cython.ccall
    def finalize(self):
        """Shrink arrays and sort per-strand coordinates in place.
        
        Each chromosome's plus- and minus-strand arrays are resized to the observed
        counts, sorted ascending, and aggregate counters such as ``total`` and
        ``length`` are refreshed. Call this after loading data.
        """
        c: bytes
        chrnames: set

        self.total = 0

        chrnames = self.get_chr_names()

        for c in chrnames:
            self.locations[c][0].resize(self.pointer[c][0], refcheck=False)
            self.locations[c][0].sort()
            self.locations[c][1].resize(self.pointer[c][1], refcheck=False)
            self.locations[c][1].sort()
            self.total += self.locations[c][0].size + self.locations[c][1].size

        self.is_sorted = True
        self.length = self.fw * self.total
        return

    @cython.ccall
    def set_rlengths(self, rlengths: dict) -> bool:
        """Attach reference chromosome lengths to the track.
        
        Parameters
        ----------
        rlengths : dict
            Mapping from chromosome name (bytes) to reference length.
        
        Returns
        -------
        bool
            True when the length mapping has been updated.
        
        Notes
        -----
        Any chromosome stored in the track but missing from ``rlengths`` is assigned
        ``INT_MAX`` so downstream bounds checks can succeed.
        """
        valid_chroms: set
        missed_chroms: set
        chrom: bytes

        valid_chroms = set(self.locations.keys()).intersection(rlengths.keys())
        for chrom in sorted(valid_chroms):
            self.rlengths[chrom] = rlengths[chrom]
        missed_chroms = set(self.locations.keys()).difference(rlengths.keys())
        for chrom in sorted(missed_chroms):
            self.rlengths[chrom] = INT_MAX
        return True

    @cython.ccall
    def get_rlengths(self) -> dict:
        """Return the reference chromosome lengths associated with the track.
        
        Returns
        -------
        dict
            Mapping from chromosome name (bytes) to reference length. Chromosomes
            without a recorded length default to ``INT_MAX``.
        """
        if not self.rlengths:
            self.rlengths = dict([(k, INT_MAX) for k in self.locations.keys()])
        return self.rlengths

    @cython.ccall
    def get_locations_by_chr(self, chromosome: bytes):
        """Return the strand-specific arrays for a chromosome.
        
        Parameters
        ----------
        chromosome : bytes
            Chromosome name, provided as bytes.
        
        Returns
        -------
        tuple[numpy.ndarray, numpy.ndarray]
            Pair of numpy arrays ``(plus, minus)`` containing 5' positions.
        
        Raises
        ------
        Exception
            If the chromosome is not present in the track.
        """
        if chromosome in self.locations:
            return self.locations[chromosome]
        else:
            raise Exception("No such chromosome name (%s) in TrackI object!\n" % (chromosome))

    @cython.ccall
    def get_chr_names(self) -> set:
        """Return a sorted set of chromosome names stored in the track.
        
        Returns
        -------
        set
            Sorted chromosome names (bytes) that currently have positions.
        """
        return set(sorted(self.locations.keys()))

    @cython.ccall
    def sort(self):
        """Sort per-strand coordinate arrays for every chromosome.
        
        Positions are ordered ascending on each strand and the ``is_sorted`` flag is
        set to ``True`` once sorting completes.
        """
        c: bytes
        chrnames: set

        chrnames = self.get_chr_names()

        for c in chrnames:
            self.locations[c][0].sort()
            self.locations[c][1].sort()

        self.is_sorted = True
        return

    @cython.boundscheck(False)  # do not check that np indices are valid
    @cython.ccall
    def filter_dup(self, maxnum: cython.int = -1) -> cython.ulong:
        """Limit duplicate 5' positions to a maximum count per strand.
        
        Parameters
        ----------
        maxnum : int, optional
            Maximum number of occurrences allowed per coordinate. A negative value
            disables duplicate filtering.
        
        Returns
        -------
        int
            Total number of retained positions across both strands after filtering.
        
        Notes
        -----
        The track must be sorted before filtering. Coordinates exceeding ``maxnum``
        are discarded, pointers are updated, and ``total``/``length`` are recomputed.
        """
        p: cython.int
        n: cython.int
        current_loc: cython.int
        # index for old array, and index for new one
        i_old: cython.ulong
        i_new: cython.ulong
        size: cython.ulong
        k: bytes
        plus: cnp.ndarray(cython.int, ndim=1)
        new_plus: cnp.ndarray(cython.int, ndim=1)
        minus: cnp.ndarray(cython.int, ndim=1)
        new_minus: cnp.ndarray(cython.int, ndim=1)
        chrnames: set

        if maxnum < 0:
            return self.total         # do nothing

        if not self.is_sorted:
            self.sort()

        self.total = 0
        self.length = 0

        chrnames = self.get_chr_names()

        for k in chrnames:
            # for each chromosome.
            # This loop body is too big, I may need to split code later...

            # + strand
            i_new = 0
            plus = self.locations[k][0]
            size = plus.shape[0]
            if len(plus) <= 1:
                new_plus = plus         # do nothing
            else:
                new_plus = np.zeros(self.pointer[k][0] + 1, dtype='i4')
                new_plus[i_new] = plus[i_new]  # first item
                i_new += 1
                # the number of tags in the current location
                n = 1
                current_loc = plus[0]
                for i_old in range(1, size):
                    p = plus[i_old]
                    if p == current_loc:
                        n += 1
                    else:
                        current_loc = p
                        n = 1
                    if n <= maxnum:
                        new_plus[i_new] = p
                        i_new += 1
                new_plus.resize(i_new, refcheck=False)
                self.total += i_new
                self.pointer[k][0] = i_new
                # free memory?
                # I know I should shrink it to 0 size directly,
                # however, on Mac OSX, it seems directly assigning 0
                # doesn't do a thing.
                plus.resize(self.buffer_size, refcheck=False)
                plus.resize(0, refcheck=False)
                # hope there would be no mem leak...

            # - strand
            i_new = 0
            minus = self.locations[k][1]
            size = minus.shape[0]
            if len(minus) <= 1:
                new_minus = minus         # do nothing
            else:
                new_minus = np.zeros(self.pointer[k][1] + 1,
                                     dtype='i4')
                new_minus[i_new] = minus[i_new]  # first item
                i_new += 1
                # the number of tags in the current location
                n = 1
                current_loc = minus[0]
                for i_old in range(1, size):
                    p = minus[i_old]
                    if p == current_loc:
                        n += 1
                    else:
                        current_loc = p
                        n = 1
                    if n <= maxnum:
                        new_minus[i_new] = p
                        i_new += 1
                new_minus.resize(i_new, refcheck=False)
                self.total += i_new
                self.pointer[k][1] = i_new
                # free memory ?
                # I know I should shrink it to 0 size directly,
                # however, on Mac OSX, it seems directly assigning 0
                # doesn't do a thing.
                minus.resize(self.buffer_size, refcheck=False)
                minus.resize(0, refcheck=False)
                # hope there would be no mem leak...

            self.locations[k] = [new_plus, new_minus]

        self.length = self.fw * self.total
        return self.total

    @cython.ccall
    def sample_percent(self, percent: cython.float, seed: cython.int = -1):
        """Down-sample positions in place by a fixed percentage.
        
        Parameters
        ----------
        percent : float
            Fraction of positions to keep per strand between 0 and 1 (inclusive).
        seed : int, optional
            Seed forwarded to NumPy's RNG; a negative value uses global state.
        
        Notes
        -----
        Sampling is performed independently for plus and minus strands by shuffling
        each array, resizing to the requested fraction, and restoring sort order.
        Aggregate counters ``total`` and ``length`` are refreshed.
        """
        num: cython.int  # num: number of reads allowed on a certain chromosome
        k: bytes
        chrnames: set

        self.total = 0
        self.length = 0

        chrnames = self.get_chr_names()

        if seed >= 0:
            np.random.seed(seed)

        for k in chrnames:
            # for each chromosome.
            # This loop body is too big, I may need to split code later...

            num = cython.cast(cython.int,
                              round(self.locations[k][0].shape[0] * percent, 5))
            np.random.shuffle(self.locations[k][0])
            self.locations[k][0].resize(num, refcheck=False)
            self.locations[k][0].sort()
            self.pointer[k][0] = self.locations[k][0].shape[0]

            num = cython.cast(cython.int,
                              round(self.locations[k][1].shape[0] * percent, 5))
            np.random.shuffle(self.locations[k][1])
            self.locations[k][1].resize(num, refcheck=False)
            self.locations[k][1].sort()
            self.pointer[k][1] = self.locations[k][1].shape[0]

            self.total += self.pointer[k][0] + self.pointer[k][1]

        self.length = self.fw * self.total
        return

    @cython.ccall
    def sample_num(self, samplesize: cython.ulong, seed: cython.int = -1):
        """Down-sample positions in place so the total approximates ``samplesize``.
        
        Parameters
        ----------
        samplesize : int
            Target number of positions across both strands.
        seed : int, optional
            Seed forwarded to :meth:`sample_percent`.
        
        Notes
        -----
        The method converts ``samplesize`` into a sampling fraction using the current
        ``total`` and reuses :meth:`sample_percent`.
        """
        percent: cython.float

        percent = cython.cast(cython.float, samplesize) / self.total
        self.sample_percent(percent, seed)
        return

    @cython.ccall
    def print_to_bed(self, fhd=None):
        """Stream the track as BED records.
        
        Parameters
        ----------
        fhd : io.IOBase, optional
            Writable file-like object. Defaults to ``sys.stdout``.
        
        Notes
        -----
        Emits one record per stored position with fixed-width intervals derived from
        ``fw`` and strand-specific orientation.
        """
        i: cython.int
        p: cython.int
        k: bytes
        chrnames: set

        if not fhd:
            fhd = sys.stdout
        assert isinstance(fhd, io.IOBase)
        assert self.fw > 0, "FWTrack object .fw should be set larger than 0!"

        chrnames = self.get_chr_names()

        for k in chrnames:
            # for each chromosome.
            # This loop body is too big, I may need to split code later...

            plus = self.locations[k][0]

            for i in range(plus.shape[0]):
                p = plus[i]
                fhd.write("%s\t%d\t%d\t.\t.\t%s\n" % (k.decode(),
                                                      p,
                                                      p + self.fw,
                                                      "+"))

            minus = self.locations[k][1]

            for i in range(minus.shape[0]):
                p = minus[i]
                fhd.write("%s\t%d\t%d\t.\t.\t%s\n" % (k.decode(),
                                                      p-self.fw,
                                                      p,
                                                      "-"))
        return

    @cython.ccall
    def extract_region_tags(self, chromosome: bytes,
                            startpos: cython.int, endpos: cython.int) -> tuple:
        """Collect positions within a genomic window for both strands.
        
        Parameters
        ----------
        chromosome : bytes
            Chromosome identifier to query.
        startpos : int
            Inclusive start coordinate of the window.
        endpos : int
            Inclusive end coordinate of the window.
        
        Returns
        -------
        tuple[numpy.ndarray, numpy.ndarray]
            Pair of numpy arrays ``(plus, minus)`` containing positions inside the
            requested window.
        
        Notes
        -----
        The track is sorted on demand before performing the windowed lookup.
        """
        i: cython.int
        pos: cython.int
        rt_plus: np.ndarray(cython.int, ndim=1)
        rt_minus: np.ndarray(cython.int, ndim=1)
        temp: list
        chrnames: set

        if not self.is_sorted:
            self.sort()

        chrnames = self.get_chr_names()
        assert chromosome in chrnames, "chromosome %s can't be found in the FWTrack object." % chromosome

        (plus, minus) = self.locations[chromosome]

        temp = []
        for i in range(plus.shape[0]):
            pos = plus[i]
            if pos < startpos:
                continue
            elif pos > endpos:
                break
            else:
                temp.append(pos)
        rt_plus = np.array(temp)

        temp = []
        for i in range(minus.shape[0]):
            pos = minus[i]
            if pos < startpos:
                continue
            elif pos > endpos:
                break
            else:
                temp.append(pos)
        rt_minus = np.array(temp)
        return (rt_plus, rt_minus)

    @cython.ccall
    def compute_region_tags_from_peaks(self, peaks: PeakIO,
                                       func,
                                       window_size: cython.int = 100,
                                       cutoff: cython.float = 5.0) -> list:
        """Apply a summary function to tags collected around peak regions.
        
        Parameters
        ----------
        peaks : MACS3.IO.PeakIO.PeakIO
            Peak container providing genomic intervals and metadata.
        func : callable
            Callback invoked as ``func(chrom, plus, minus, startpos, endpos, ...)``
            for each peak. The callable must accept ``window_size`` and ``cutoff``
            keyword arguments.
        window_size : int, optional
            Half-window size added on each side of every peak when collecting tags.
        cutoff : float, optional
            Additional threshold passed to ``func``.
        
        Returns
        -------
        list
            Results returned by ``func`` for each processed peak.
        
        Notes
        -----
        Both the track and the ``peaks`` object are sorted before iteration, and
        per-chromosome state is reused to avoid rescanning arrays.
        """
        m: cython.int
        i: cython.int
        j: cython.int
        pos: cython.int
        startpos: cython.int
        endpos: cython.int

        plus: cnp.ndarray(cython.int, ndim=1)
        minus: cnp.ndarray(cython.int, ndim=1)
        rt_plus: cnp.ndarray(cython.int, ndim=1)
        rt_minus: cnp.ndarray(cython.int, ndim=1)

        chrom: bytes
        name: bytes

        temp: list
        retval: list
        pchrnames: set
        chrnames: set

        pchrnames = peaks.get_chr_names()
        retval = []

        # this object should be sorted
        if not self.is_sorted:
            self.sort()
        # PeakIO object should be sorted
        peaks.sort()

        chrnames = self.get_chr_names()

        for chrom in sorted(pchrnames):
            assert chrom in chrnames, "chromosome %s can't be found in the FWTrack object." % chrom
            (plus, minus) = self.locations[chrom]
            cpeaks = peaks.get_data_from_chrom(chrom)
            prev_i = 0
            prev_j = 0
            for m in range(len(cpeaks)):
                startpos = cpeaks[m]["start"] - window_size
                endpos = cpeaks[m]["end"] + window_size
                name = cpeaks[m]["name"]

                temp = []
                for i in range(prev_i, plus.shape[0]):
                    pos = plus[i]
                    if pos < startpos:
                        continue
                    elif pos > endpos:
                        prev_i = i
                        break
                    else:
                        temp.append(pos)
                rt_plus = np.array(temp, dtype="i4")

                temp = []
                for j in range(prev_j, minus.shape[0]):
                    pos = minus[j]
                    if pos < startpos:
                        continue
                    elif pos > endpos:
                        prev_j = j
                        break
                    else:
                        temp.append(pos)
                rt_minus = np.array(temp, dtype="i4")

                retval.append(func(chrom, rt_plus, rt_minus, startpos, endpos,
                                   name=name,
                                   window_size=window_size,
                                   cutoff=cutoff))
                # rewind window_size
                for i in range(prev_i, 0, -1):
                    if plus[prev_i] - plus[i] >= window_size:
                        break
                prev_i = i

                for j in range(prev_j, 0, -1):
                    if minus[prev_j] - minus[j] >= window_size:
                        break
                prev_j = j
                # end of a loop

        return retval

    @cython.ccall
    def pileup_a_chromosome(self, chrom: bytes,
                            d: cython.long,
                            scale_factor: cython.float = 1.0,
                            baseline_value: cython.float = 0.0,
                            directional: bool = True,
                            end_shift: cython.int = 0) -> list:
        """Compute a coverage pileup for a single chromosome.
        
        Parameters
        ----------
        chrom : bytes
            Chromosome name to pile up.
        d : int
            Extension length applied in the 3' direction unless ``directional`` is
            ``False``.
        scale_factor : float, optional
            Value used to scale the resulting coverage.
        baseline_value : float, optional
            Minimum value enforced on the coverage array.
        directional : bool, optional
            If ``False``, extend cuts symmetrically to both sides by ``d / 2``.
        end_shift : int, optional
            Shift applied to the 5' cuts before extension; positive values move
            toward the 3' direction.
        
        Returns
        -------
        list
            Two-element list ``[positions, values]`` with numpy arrays describing
            the pileup breakpoints and scaled coverage.
        """
        five_shift: cython.long
        # adjustment to 5' end and 3' end positions to make a fragment
        three_shift: cython.long
        rlength: cython.long
        chrlengths: dict
        tmp_pileup: list

        chrlengths = self.get_rlengths()
        rlength = chrlengths[chrom]

        # adjust extension length according to 'directional' and
        # 'halfextension' setting.
        if directional:
            # only extend to 3' side
            five_shift = - end_shift
            three_shift = end_shift + d
        else:
            # both sides
            five_shift = d//2 - end_shift
            three_shift = end_shift + d - d//2

        tmp_pileup = se_all_in_one_pileup(self.locations[chrom][0],
                                          self.locations[chrom][1],
                                          five_shift,
                                          three_shift,
                                          rlength,
                                          scale_factor,
                                          baseline_value)
        return tmp_pileup

    @cython.ccall
    def pileup_a_chromosome_c(self, chrom: bytes, ds: list,
                              scale_factor_s: list,
                              baseline_value: cython.float = 0.0,
                              directional: bool = True,
                              end_shift: cython.int = 0) -> list:
        """Compute a control pileup using multiple extension lengths.
        
        Parameters
        ----------
        chrom : bytes
            Chromosome name to pile up.
        ds : list[int]
            Extension lengths used to build individual pileups.
        scale_factor_s : list[float]
            Scale factors paired with each entry in ``ds``.
        baseline_value : float, optional
            Minimum value enforced on the coverage array.
        directional : bool, optional
            If ``False``, extend cuts symmetrically to both sides by ``d / 2``.
        end_shift : int, optional
            Shift applied to the 5' cuts before extension; positive values move
            toward the 3' direction.
        
        Returns
        -------
        list
            Two-element list ``[positions, values]`` representing the merged pileup
            where the maximum value is taken across the supplied extensions.
        """
        d: cython.long
        five_shift: cython.long
        # adjustment to 5' end and 3' end positions to make a fragment
        three_shift: cython.long
        rlength: cython.long
        chrlengths: dict
        five_shift_s: list = []
        three_shift_s: list = []
        tmp_pileup: list
        prev_pileup: list

        chrlengths = self.get_rlengths()
        rlength = chrlengths[chrom]
        assert len(ds) == len(scale_factor_s), "ds and scale_factor_s must have the same length!"

        # adjust extension length according to 'directional' and
        # 'halfextension' setting.
        for d in ds:
            if directional:
                # only extend to 3' side
                five_shift_s.append(- end_shift)
                three_shift_s.append(end_shift + d)
            else:
                # both sides
                five_shift_s.append(d//2 - end_shift)
                three_shift_s.append(end_shift + d - d//2)

        prev_pileup = None

        for i in range(len(ds)):
            five_shift = five_shift_s[i]
            three_shift = three_shift_s[i]
            scale_factor = scale_factor_s[i]
            tmp_pileup = se_all_in_one_pileup(self.locations[chrom][0],
                                              self.locations[chrom][1],
                                              five_shift,
                                              three_shift,
                                              rlength,
                                              scale_factor,
                                              baseline_value)

            if prev_pileup:
                prev_pileup = over_two_pv_array(prev_pileup,
                                                tmp_pileup,
                                                func="max")
            else:
                prev_pileup = tmp_pileup

        return prev_pileup


@cython.inline
@cython.cfunc
def left_sum(data,
             pos: cython.int,
             width: cython.int) -> cython.int:
    return sum([data[x] for x in data if x <= pos and x >= pos - width])


@cython.inline
@cython.cfunc
def right_sum(data,
              pos: cython.int,
              width: cython.int) -> cython.int:
    return sum([data[x] for x in data if x >= pos and x <= pos + width])


@cython.inline
@cython.cfunc
def left_forward(data,
                 pos: cython.int,
                 window_size: cython.int) -> cython.int:
    return data.get(pos, 0) - data.get(pos-window_size, 0)


@cython.inline
@cython.cfunc
def right_forward(data,
                  pos: cython.int,
                  window_size: cython.int) -> cython.int:
    return data.get(pos + window_size, 0) - data.get(pos, 0)



# Time-stamp: <2024-10-02 16:06:33 Tao Liu>

"""Description: compare bdg files

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

import os
from MACS3.IO import BedGraphIO
from MACS3.Utilities.OptValidator import opt_validate_bdgcmp

# ------------------------------------
# constants
# ------------------------------------

# ------------------------------------
# Misc functions
# ------------------------------------

# ------------------------------------
# Main function
# ------------------------------------


def run(options):
    options = opt_validate_bdgcmp(options)
    info = options.info
    # warn = options.warn
    # debug = options.debug
    # error = options.error

    scaling_factor = options.sfactor

    # not an actual depth, but its reciprocal, a trick to override
    # SPMR while necessary.
    pseudo_depth = 1.0/scaling_factor

    info("Read and build treatment bedGraph...")
    tbio = BedGraphIO.bedGraphIO(options.tfile)
    tbtrack = tbio.read_bedGraph()

    info("Read and build control bedGraph...")
    cbio = BedGraphIO.bedGraphIO(options.cfile)
    cbtrack = cbio.read_bedGraph()

    info("Build ScoreTrackII...")
    sbtrack = tbtrack.make_ScoreTrackII_for_macs(cbtrack, depth1=pseudo_depth, depth2=pseudo_depth)
    if abs(scaling_factor-1) > 1e-6:
        # Only for the case while your input is SPMR from MACS3 callpeak; Let's override SPMR.
        info("Values in your input bedGraph files will be multiplied by %f ..." % scaling_factor)
        sbtrack.change_normalization_method(ord('M'))  # a hack to override SPMR
    sbtrack.set_pseudocount(options.pseudocount)

    already_processed_method_list = []
    for (i, method) in enumerate(options.method):
        if method in already_processed_method_list:
            continue
        else:
            already_processed_method_list.append(method)

        info("Calculate scores comparing treatment and control by '%s'..." % method)
        if options.ofile:
            ofile = os.path.join(options.outdir, options.ofile[i])
        else:
            ofile = os.path.join(options.outdir, options.oprefix + "_" + method + ".bdg")
        # build score track
        if method == 'ppois':
            sbtrack.change_score_method(ord('p'))
        elif method == 'qpois':
            sbtrack.change_score_method(ord('q'))
        elif method == 'subtract':
            sbtrack.change_score_method(ord('d'))
        elif method == 'logFE':
            sbtrack.change_score_method(ord('f'))
        elif method == 'FE':
            sbtrack.change_score_method(ord('F'))
        elif method == 'logLR':             # log likelihood
            sbtrack.change_score_method(ord('l'))
        elif method == 'slogLR':             # log likelihood
            sbtrack.change_score_method(ord('s'))
        elif method == 'max':
            sbtrack.change_score_method(ord('M'))
        else:
            raise Exception("Can't reach here!")

        info("Write bedGraph of scores...")
        ofhd = open(ofile, "w")
        # write_bedGraph function for ScoreTrack
        sbtrack.write_bedGraph(ofhd, name="%s_Scores" % (method.upper()), description="Scores calculated by %s" % (method.upper()), column=3)
        info("Finished '%s'! Please check '%s'!" % (method, ofile))
